from __future__ import annotations

import json
import logging
from collections.abc import Coroutine, Sequence
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar, Union

import click

from chik.cmds import options
from chik.cmds.param_types import Bytes32ParamType
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint64

_T = TypeVar("_T")

FC = TypeVar("FC", bound=Union[Callable[..., Any], click.Command])

logger = logging.getLogger(__name__)


# TODO: this is more general and should be part of refactoring the overall CLI code duplication
def run(coro: Coroutine[Any, Any, Optional[dict[str, Any]]]) -> None:
    import asyncio

    response = asyncio.run(coro)

    success = response is not None and response.get("success", False)
    logger.info(f"data layer cli call response:{success}")
    # todo make sure all cli methods follow this pattern, uncomment
    # if not success:
    # raise click.ClickException(message=f"query unsuccessful, response: {response}")


@click.group("data", help="Manage your data")
def data_cmd() -> None:
    pass


# TODO: maybe use more helpful `type=`s to get click to handle error reporting of
#       malformed inputs.


def create_changelist_option() -> Callable[[FC], FC]:
    return click.option(
        "-d",
        "--changelist",
        "changelist_string",
        help="str representing the changelist",
        type=str,
        required=True,
    )


def create_store_updates_option() -> Callable[[FC], FC]:
    return click.option(
        "-d",
        "--store_updates",
        "store_updates_string",
        help="str representing the store updates",
        type=str,
        required=True,
    )


def create_key_option(multiple: bool = False) -> Callable[[FC], FC]:
    return click.option(
        "-k",
        "--key",
        "key_strings" if multiple else "key_string",
        help="str representing the key",
        type=str,
        required=True,
        multiple=multiple,
    )


def create_data_store_id_option() -> Callable[[FC], FC]:
    return click.option(
        "-store",
        "--id",
        help="The hexadecimal store id.",
        type=Bytes32ParamType(),
        required=True,
    )


def create_data_store_name_option() -> Callable[[FC], FC]:
    return click.option(
        "-n",
        "--table_name",
        "table_name",
        help="The name of the table.",
        type=str,
        required=True,
    )


def create_rpc_port_option() -> Callable[[FC], FC]:
    return click.option(
        "-dp",
        "--data-rpc-port",
        help="Set the port where the data layer is hosting the RPC interface. See rpc_port under wallet in config.yaml",
        type=int,
        default=None,
        show_default=True,
    )


def create_root_hash_option() -> Callable[[FC], FC]:
    return click.option(
        "-r",
        "--root_hash",
        help="The hexadecimal root hash",
        type=Bytes32ParamType(),
        required=False,
    )


def create_page_option() -> Callable[[FC], FC]:
    return click.option(
        "-p",
        "--page",
        help="Enables pagination of the output and requests a specific page.",
        type=int,
        required=False,
    )


def create_max_page_size_option() -> Callable[[FC], FC]:
    return click.option(
        "--max-page-size",
        help="Set how many bytes to be included in a page, if pagination is enabled.",
        type=int,
        required=False,
    )


# Functions with this mark in this file are not being ported to @tx_out_cmd due to API peculiarities
# They will therefore not work with observer-only functionality
# NOTE: tx_endpoint  (This creates wallet transactions and should be parametrized by relevant options)
@data_cmd.command("create_data_store", help="Create a new data store")
@create_rpc_port_option()
@options.create_fee()
@click.option("--verbose", is_flag=True, help="Enable verbose output.")
@options.create_fingerprint()
def create_data_store(
    data_rpc_port: int,
    fee: Optional[uint64],
    verbose: bool,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import create_data_store_cmd

    run(create_data_store_cmd(data_rpc_port, fee, verbose, fingerprint=fingerprint))


@data_cmd.command("get_value", help="Get the value for a given key and store")
@create_data_store_id_option()
@create_key_option()
@create_root_hash_option()
@create_rpc_port_option()
@options.create_fingerprint()
def get_value(
    id: bytes32,
    key_string: str,
    root_hash: Optional[bytes32],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_value_cmd

    run(get_value_cmd(data_rpc_port, id, key_string, root_hash, fingerprint=fingerprint))


# NOTE: tx_endpoint
@data_cmd.command("update_data_store", help="Update a store by providing the changelist operations")
@create_data_store_id_option()
@create_changelist_option()
@create_rpc_port_option()
@options.create_fee()
@options.create_fingerprint()
@click.option("--submit/--no-submit", default=True, help="Submit the result on chain")
def update_data_store(
    id: bytes32,
    changelist_string: str,
    data_rpc_port: int,
    fee: Optional[uint64],
    fingerprint: Optional[int],
    submit: bool,
) -> None:
    from chik.cmds.data_funcs import update_data_store_cmd

    run(
        update_data_store_cmd(
            rpc_port=data_rpc_port,
            store_id=id,
            changelist=json.loads(changelist_string),
            fee=fee,
            fingerprint=fingerprint,
            submit_on_chain=submit,
        )
    )


@data_cmd.command("update_multiple_stores", help="Update multiple stores by providing the changelist operations")
@create_store_updates_option()
@create_rpc_port_option()
@options.create_fee()
@options.create_fingerprint()
@click.option("--submit/--no-submit", default=True, help="Submit the result on chain")
def update_multiple_stores(
    store_updates_string: str,
    data_rpc_port: int,
    fee: uint64,
    fingerprint: Optional[int],
    submit: bool,
) -> None:
    from chik.cmds.data_funcs import update_multiple_stores_cmd

    run(
        update_multiple_stores_cmd(
            rpc_port=data_rpc_port,
            store_updates=json.loads(store_updates_string),
            fee=fee,
            fingerprint=fingerprint,
            submit_on_chain=submit,
        )
    )


@data_cmd.command("submit_pending_root", help="Submit on chain a locally stored batch")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fee()
@options.create_fingerprint()
def submit_pending_root(
    id: bytes32,
    data_rpc_port: int,
    fee: uint64,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import submit_pending_root_cmd

    run(
        submit_pending_root_cmd(
            rpc_port=data_rpc_port,
            store_id=id,
            fee=fee,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("submit_all_pending_roots", help="Submit on chain all locally stored batches")
@create_rpc_port_option()
@options.create_fee()
@options.create_fingerprint()
def submit_all_pending_roots(
    data_rpc_port: int,
    fee: uint64,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import submit_all_pending_roots_cmd

    run(
        submit_all_pending_roots_cmd(
            rpc_port=data_rpc_port,
            fee=fee,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("get_keys", help="Get all keys for a given store")
@create_data_store_id_option()
@create_root_hash_option()
@create_rpc_port_option()
@options.create_fingerprint()
@create_page_option()
@create_max_page_size_option()
def get_keys(
    id: bytes32,
    root_hash: Optional[bytes32],
    data_rpc_port: int,
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_keys_cmd

    run(get_keys_cmd(data_rpc_port, id, root_hash, fingerprint=fingerprint, page=page, max_page_size=max_page_size))


@data_cmd.command("get_keys_values", help="Get all keys and values for a given store")
@create_data_store_id_option()
@create_root_hash_option()
@create_rpc_port_option()
@options.create_fingerprint()
@create_page_option()
@create_max_page_size_option()
def get_keys_values(
    id: bytes32,
    root_hash: Optional[bytes32],
    data_rpc_port: int,
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_keys_values_cmd

    run(
        get_keys_values_cmd(
            data_rpc_port, id, root_hash, fingerprint=fingerprint, page=page, max_page_size=max_page_size
        )
    )


@data_cmd.command("get_root", help="Get the published root hash value for a given store")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fingerprint()
def get_root(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_root_cmd

    run(get_root_cmd(rpc_port=data_rpc_port, store_id=id, fingerprint=fingerprint))


@data_cmd.command("subscribe", help="Subscribe to a store")
@create_data_store_id_option()
@click.option(
    "-u",
    "--url",
    "urls",
    help="Manually provide a list of servers urls for downloading the data",
    type=str,
    multiple=True,
)
@create_rpc_port_option()
@options.create_fingerprint()
def subscribe(
    id: bytes32,
    urls: list[str],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import subscribe_cmd

    run(subscribe_cmd(rpc_port=data_rpc_port, store_id=id, urls=urls, fingerprint=fingerprint))


@data_cmd.command("remove_subscription", help="Remove server urls that are added via subscribing to urls")
@create_data_store_id_option()
@click.option("-u", "--url", "urls", help="Server urls to remove", type=str, multiple=True)
@create_rpc_port_option()
@options.create_fingerprint()
def remove_subscription(
    id: bytes32,
    urls: list[str],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import remove_subscriptions_cmd

    run(remove_subscriptions_cmd(rpc_port=data_rpc_port, store_id=id, urls=urls, fingerprint=fingerprint))


@data_cmd.command("unsubscribe", help="Completely untrack a store")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fingerprint()
@click.option("--retain", is_flag=True, help="Retain .dat files")
def unsubscribe(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
    retain: bool,
) -> None:
    from chik.cmds.data_funcs import unsubscribe_cmd

    run(unsubscribe_cmd(rpc_port=data_rpc_port, store_id=id, fingerprint=fingerprint, retain=retain))


@data_cmd.command(
    "get_kv_diff", help="Get the inserted and deleted keys and values between an initial and a final hash"
)
@create_data_store_id_option()
@click.option("-hash_1", "--hash_1", help="Initial hash", type=Bytes32ParamType(), required=True)
@click.option("-hash_2", "--hash_2", help="Final hash", type=Bytes32ParamType(), required=True)
@create_rpc_port_option()
@options.create_fingerprint()
@create_page_option()
@create_max_page_size_option()
def get_kv_diff(
    id: bytes32,
    hash_1: bytes32,
    hash_2: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_kv_diff_cmd

    run(
        get_kv_diff_cmd(
            rpc_port=data_rpc_port,
            store_id=id,
            hash_1=hash_1,
            hash_2=hash_2,
            fingerprint=fingerprint,
            page=page,
            max_page_size=max_page_size,
        )
    )


@data_cmd.command("get_root_history", help="Get all changes of a singleton")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fingerprint()
def get_root_history(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_root_history_cmd

    run(get_root_history_cmd(rpc_port=data_rpc_port, store_id=id, fingerprint=fingerprint))


@data_cmd.command("add_missing_files", help="Manually reconstruct server files from the data layer database")
@click.option(
    "-i",
    "--ids",
    help="List of stores to reconstruct. If not specified, all stores will be reconstructed",
    type=str,
    multiple=True,
    required=False,
)
@click.option(
    "-o/-n",
    "--overwrite/--no-overwrite",
    help="Specify if already existing files need to be overwritten by this command",
)
@click.option(
    "-d", "--directory", type=str, help="If specified, use a non-default directory to write the files", required=False
)
@create_rpc_port_option()
@options.create_fingerprint()
def add_missing_files(
    ids: Sequence[bytes32],
    overwrite: bool,
    directory: Optional[str],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import add_missing_files_cmd

    run(
        add_missing_files_cmd(
            rpc_port=data_rpc_port,
            ids=list(ids) if ids else None,
            overwrite=overwrite,
            foldername=None if directory is None else Path(directory),
            fingerprint=fingerprint,
        )
    )


# NOTE: tx_endpoint
@data_cmd.command("add_mirror", help="Publish mirror urls on chain")
@create_data_store_id_option()
@click.option(
    "-a", "--amount", help="Amount to spend for this mirror, in mojos", type=int, default=0, show_default=True
)
@click.option(
    "-u",
    "--url",
    "urls",
    help="URL to publish on the new coin, multiple accepted and will be published to a single coin.",
    type=str,
    multiple=True,
)
@options.create_fee()
@create_rpc_port_option()
@options.create_fingerprint()
def add_mirror(
    id: bytes32,
    amount: int,
    urls: list[str],
    fee: Optional[uint64],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import add_mirror_cmd

    run(
        add_mirror_cmd(
            rpc_port=data_rpc_port,
            store_id=id,
            urls=urls,
            amount=amount,
            fee=fee,
            fingerprint=fingerprint,
        )
    )


# NOTE: tx_endpoint
@data_cmd.command("delete_mirror", help="Delete an owned mirror by its coin id")
@click.option("-c", "--coin_id", help="Coin id", type=Bytes32ParamType(), required=True)
@options.create_fee()
@create_rpc_port_option()
@options.create_fingerprint()
def delete_mirror(
    coin_id: bytes32,
    fee: Optional[uint64],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import delete_mirror_cmd

    run(
        delete_mirror_cmd(
            rpc_port=data_rpc_port,
            coin_id=coin_id,
            fee=fee,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("get_mirrors", help="Get a list of all mirrors for a given store")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fingerprint()
def get_mirrors(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_mirrors_cmd

    run(
        get_mirrors_cmd(
            rpc_port=data_rpc_port,
            store_id=id,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("get_subscriptions", help="Get subscribed stores, including the owned stores")
@create_rpc_port_option()
@options.create_fingerprint()
def get_subscriptions(
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_subscriptions_cmd

    run(
        get_subscriptions_cmd(
            rpc_port=data_rpc_port,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("get_owned_stores", help="Get owned stores")
@create_rpc_port_option()
@options.create_fingerprint()
def get_owned_stores(
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_owned_stores_cmd

    run(
        get_owned_stores_cmd(
            rpc_port=data_rpc_port,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command("get_sync_status", help="Get locally stored root compared to the root of the singleton")
@create_data_store_id_option()
@create_rpc_port_option()
@options.create_fingerprint()
def get_sync_status(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_sync_status_cmd

    run(get_sync_status_cmd(rpc_port=data_rpc_port, store_id=id, fingerprint=fingerprint))


@data_cmd.group("plugins", help="Get information about configured uploader/downloader plugins")
def plugins_cmd() -> None:
    pass


@plugins_cmd.command("check", help="Calls the plugin_info endpoint on all configured plugins")
@create_rpc_port_option()
def check_plugins(
    data_rpc_port: int,
) -> None:
    from chik.cmds.data_funcs import check_plugins_cmd

    run(check_plugins_cmd(rpc_port=data_rpc_port))


@data_cmd.command(
    "clear_pending_roots",
    help="Clear pending roots that will not be published, associated data may not be recoverable",
)
@create_data_store_id_option()
@click.confirmation_option(
    prompt="Associated data may not be recoverable.\nAre you sure you want to remove the pending roots?",
)
@create_rpc_port_option()
@options.create_fingerprint()
def clear_pending_roots(
    id: bytes32,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import clear_pending_roots

    run(
        clear_pending_roots(
            rpc_port=data_rpc_port,
            store_id=id,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command(
    "wallet_log_in",
    help="Request that the wallet service be logged in to the specified fingerprint",
)
@create_rpc_port_option()
@options.create_fingerprint(required=True)
def wallet_log_in(
    data_rpc_port: int,
    fingerprint: int,
) -> None:
    from chik.cmds.data_funcs import wallet_log_in_cmd

    run(
        wallet_log_in_cmd(
            rpc_port=data_rpc_port,
            fingerprint=fingerprint,
        )
    )


@data_cmd.command(
    "get_proof",
    help="Obtains a merkle proof of inclusion for a given key",
)
@create_data_store_id_option()
@create_rpc_port_option()
@create_key_option(multiple=True)
@options.create_fingerprint()
def get_proof(
    id: bytes32,
    key_strings: list[str],
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import get_proof_cmd

    run(get_proof_cmd(rpc_port=data_rpc_port, store_id=id, fingerprint=fingerprint, key_strings=key_strings))


@data_cmd.command(
    "verify_proof",
    help="Verifies a merkle proof of inclusion",
)
@click.option(
    "-p",
    "--proof",
    "proof_string",
    help="Proof to validate in JSON format.",
    type=str,
)
@create_rpc_port_option()
@options.create_fingerprint()
def verify_proof(
    proof_string: str,
    data_rpc_port: int,
    fingerprint: Optional[int],
) -> None:
    from chik.cmds.data_funcs import verify_proof_cmd

    proof_dict = json.loads(proof_string)
    run(verify_proof_cmd(rpc_port=data_rpc_port, fingerprint=fingerprint, proof=proof_dict))



from __future__ import annotations

from chikvdf import prove

from chik.consensus.constants import ConsensusConstants
from chik.types.blockchain_format.classgroup import ClassgroupElement
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.blockchain_format.vdf import VDFInfo, VDFProof
from chik.util.ints import uint8, uint64


def get_vdf_info_and_proof(
    constants: ConsensusConstants,
    vdf_input: ClassgroupElement,
    challenge_hash: bytes32,
    number_iters: uint64,
    normalized_to_identity: bool = False,
) -> tuple[VDFInfo, VDFProof]:
    form_size = ClassgroupElement.get_size()
    result: bytes = prove(
        bytes(challenge_hash),
        vdf_input.data,
        constants.DISCRIMINANT_SIZE_BITS,
        number_iters,
        "",
    )

    output = ClassgroupElement.create(result[:form_size])
    proof_bytes = result[form_size : 2 * form_size]
    return VDFInfo(challenge_hash, number_iters, output), VDFProof(uint8(0), proof_bytes, normalized_to_identity)



from __future__ import annotations

import logging
from typing import Any

from typing_extensions import Protocol

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.wallet.payment import Payment
from chik.wallet.puzzles.clawback.puzzle_decorator import ClawbackPuzzleDecorator
from chik.wallet.util.puzzle_decorator_type import PuzzleDecoratorType


class PuzzleDecoratorProtocol(Protocol):
    @staticmethod
    def create(config: dict[str, Any]) -> PuzzleDecoratorProtocol: ...

    def decorate(self, inner_puzzle: Program) -> Program: ...

    def decorate_target_puzzle_hash(
        self, inner_puzzle: Program, target_puzzle_hash: bytes32
    ) -> tuple[Program, bytes32]: ...

    def decorate_memos(
        self, inner_puzzle: Program, target_puzzle_hash: bytes32, memos: list[bytes]
    ) -> tuple[Program, list[bytes]]: ...

    def solve(self, puzzle: Program, primaries: list[Payment], inner_solution: Program) -> tuple[Program, Program]: ...


class PuzzleDecoratorManager:
    decorator_list: list[PuzzleDecoratorProtocol]
    log: logging.Logger

    @staticmethod
    def create(config: list[dict[str, Any]]) -> PuzzleDecoratorManager:
        """
        Create a new puzzle decorator manager
        :param config: Config
        :return:
        """
        self = PuzzleDecoratorManager()
        self.log = logging.getLogger(__name__)
        self.decorator_list = []
        for decorator in config:
            if "decorator" not in decorator:
                logging.error(f"Undefined decorator: {decorator}")
                continue
            decorator_name = decorator["decorator"]
            if decorator_name == PuzzleDecoratorType.CLAWBACK.name:
                self.decorator_list.append(ClawbackPuzzleDecorator.create(decorator))
            else:
                logging.error(f"Unknown puzzle decorator type: {decorator}")
        return self

    def decorate(self, inner_puzzle: Program) -> Program:
        """
        Decorator a puzzle
        :param inner_puzzle: Inner puzzle
        :return: Decorated inner puzzle
        """
        for decorator in self.decorator_list:
            inner_puzzle = decorator.decorate(inner_puzzle)
        return inner_puzzle

    def decorate_target_puzzle_hash(self, inner_puzzle: Program, target_puzzle_hash: bytes32) -> bytes32:
        """
        Decorate a target puzzle hash
        :param target_puzzle_hash: Target puzzle hash
        :param inner_puzzle: Inner puzzle
        :return: Decorated target puzzle hash
        """
        for decorator in self.decorator_list:
            inner_puzzle, target_puzzle_hash = decorator.decorate_target_puzzle_hash(inner_puzzle, target_puzzle_hash)
        return target_puzzle_hash

    def solve(self, inner_puzzle: Program, primaries: list[Payment], inner_solution: Program) -> Program:
        """
        Generate the solution of the puzzle
        :param inner_puzzle: Inner puzzle
        :param primaries: Primaries list
        :param inner_solution: Solution of the inner puzzle
        :return: Decorated inner puzzle solution
        """
        for decorator in self.decorator_list:
            inner_puzzle, inner_solution = decorator.solve(inner_puzzle, primaries, inner_solution)
        return inner_solution

    def decorate_memos(self, inner_puzzle: Program, target_puzzle_hash: bytes32, memos: list[bytes]) -> list[bytes]:
        """
        Decorate a memo list
        :param inner_puzzle: Inner puzzle
        :param target_puzzle_hash: Target puzzle hash
        :param memos: memo list
        :return: Decorated memo
        """
        for decorator in self.decorator_list:
            inner_puzzle, memos = decorator.decorate_memos(inner_puzzle, target_puzzle_hash, memos)
        return memos



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from chik.protocols.wallet_protocol import CoinState
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32, uint64
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class CoinRecord(Streamable):
    """
    These are values that correspond to a CoinName that are used
    in keeping track of the unspent database.
    """

    coin: Coin
    confirmed_block_index: uint32
    spent_block_index: uint32
    coinbase: bool
    timestamp: uint64  # Timestamp of the block at height confirmed_block_index

    @property
    def spent(self) -> bool:
        return self.spent_block_index > 0

    @property
    def name(self) -> bytes32:
        return self.coin.name()

    @property
    def coin_state(self) -> CoinState:
        spent_h = None
        if self.spent:
            spent_h = self.spent_block_index
        confirmed_height: Optional[uint32] = self.confirmed_block_index
        if self.confirmed_block_index == 0 and self.timestamp == 0:
            confirmed_height = None
        return CoinState(self.coin, spent_h, confirmed_height)



from __future__ import annotations

import asyncio
import logging
import sys
from pathlib import Path

import click

from chik.cmds.cmd_classes import ChikCliContext
from chik.plotting.util import add_plot_directory, validate_plot_size

log = logging.getLogger(__name__)


def show_plots(root_path: Path) -> None:
    from chik.plotting.util import get_plot_directories

    print("Directories where plots are being searched for:")
    print("Note that subdirectories must be added manually")
    print(
        "Add with 'chik plots add -d [dir]' and remove with"
        + " 'chik plots remove -d [dir]'"
        + " Scan and check plots with 'chik plots check'"
    )
    print()
    for str_path in get_plot_directories(root_path):
        print(f"{str_path}")


@click.group("plots", help="Manage your plots")
@click.pass_context
def plots_cmd(ctx: click.Context) -> None:
    """Create, add, remove and check your plots"""
    from chik.util.chik_logging import initialize_logging

    root_path = ChikCliContext.set_default(ctx).root_path
    if not root_path.is_dir():
        raise RuntimeError("Please initialize (or migrate) your config directory with 'chik init'")
    initialize_logging("", {"log_level": "INFO", "log_stdout": True}, root_path)


@plots_cmd.command("create", help="Create plots")
@click.option("-k", "--size", help="Plot size", type=int, default=32, show_default=True)
@click.option("--override-k", help="Force size smaller than 32", default=False, show_default=True, is_flag=True)
@click.option("-n", "--num", help="Number of plots or challenges", type=int, default=1, show_default=True)
@click.option("-b", "--buffer", help="Megabytes for sort/plot buffer", type=int, default=3389, show_default=True)
@click.option("-r", "--num_threads", help="Number of threads to use", type=int, default=2, show_default=True)
@click.option("-u", "--buckets", help="Number of buckets", type=int, default=128, show_default=True)
@click.option(
    "-a",
    "--alt_fingerprint",
    type=int,
    default=None,
    help="Enter the alternative fingerprint of the key you want to use",
)
@click.option(
    "-c",
    "--pool_contract_address",
    type=str,
    default=None,
    help="Address of where the pool reward will be sent to. Only used if alt_fingerprint and pool public key are None",
)
@click.option("-f", "--farmer_public_key", help="Hex farmer public key", type=str, default=None)
@click.option("-p", "--pool_public_key", help="Hex public key of pool", type=str, default=None)
@click.option(
    "-t",
    "--tmp_dir",
    help="Temporary directory for plotting files",
    type=click.Path(),
    default=Path("."),
    show_default=True,
)
@click.option("-2", "--tmp2_dir", help="Second temporary directory for plotting files", type=click.Path(), default=None)
@click.option(
    "-d",
    "--final_dir",
    help="Final directory for plots (relative or absolute)",
    type=click.Path(),
    default=Path("."),
    show_default=True,
)
@click.option("-i", "--plotid", help="PlotID in hex for reproducing plots (debugging only)", type=str, default=None)
@click.option("-m", "--memo", help="Memo in hex for reproducing plots (debugging only)", type=str, default=None)
@click.option("-e", "--nobitfield", help="Disable bitfield", default=False, is_flag=True)
@click.option(
    "-x", "--exclude_final_dir", help="Skips adding [final dir] to harvester for farming", default=False, is_flag=True
)
@click.option(
    "-D",
    "--connect_to_daemon",
    help="Connects to the daemon for keychain operations",
    default=False,
    is_flag=True,
    hidden=True,  # -D is only set when launched by the daemon
)
@click.pass_context
def create_cmd(
    ctx: click.Context,
    size: int,
    override_k: bool,
    num: int,
    buffer: int,
    num_threads: int,
    buckets: int,
    alt_fingerprint: int,
    pool_contract_address: str,
    farmer_public_key: str,
    pool_public_key: str,
    tmp_dir: str,
    tmp2_dir: str,
    final_dir: str,
    plotid: str,
    memo: str,
    nobitfield: bool,
    exclude_final_dir: bool,
    connect_to_daemon: bool,
) -> None:
    from chik.plotting.create_plots import create_plots, resolve_plot_keys
    from chik.plotting.util import Params

    params = Params(
        size=size,
        num=num,
        buffer=buffer,
        num_threads=num_threads,
        buckets=buckets,
        tmp_dir=Path(tmp_dir),
        tmp2_dir=Path(tmp2_dir) if tmp2_dir else None,
        final_dir=Path(final_dir),
        plotid=plotid,
        memo=memo,
        nobitfield=nobitfield,
    )

    root_path = ChikCliContext.set_default(ctx).root_path
    try:
        validate_plot_size(root_path, size, override_k)
    except ValueError as e:
        print(e)
        sys.exit(1)

    plot_keys = asyncio.run(
        resolve_plot_keys(
            farmer_public_key,
            alt_fingerprint,
            pool_public_key,
            pool_contract_address,
            root_path,
            log,
            connect_to_daemon,
        )
    )

    asyncio.run(create_plots(params, plot_keys))
    if not exclude_final_dir:
        try:
            add_plot_directory(root_path, final_dir)
        except ValueError as e:
            print(e)


@plots_cmd.command("check", help="Checks plots")
@click.option("-n", "--num", help="Number of plots or challenges", type=int, default=None)
@click.option(
    "-g",
    "--grep_string",
    help="Shows only plots that contain the string in the filename or directory name",
    type=str,
    default=None,
)
@click.option("-l", "--list_duplicates", help="List plots with duplicate IDs", default=False, is_flag=True)
@click.option("--debug-show-memo", help="Shows memo to recreate the same exact plot", default=False, is_flag=True)
@click.option("--challenge-start", help="Begins at a different [start] for -n [challenges]", type=int, default=None)
@click.pass_context
def check_cmd(
    ctx: click.Context, num: int, grep_string: str, list_duplicates: bool, debug_show_memo: bool, challenge_start: int
) -> None:
    from chik.plotting.check_plots import check_plots

    check_plots(
        ChikCliContext.set_default(ctx).root_path,
        num,
        challenge_start,
        grep_string,
        list_duplicates,
        debug_show_memo,
    )


@plots_cmd.command("add", help="Adds a directory of plots")
@click.option(
    "-d",
    "--final_dir",
    help="Final directory for plots (relative or absolute)",
    type=click.Path(),
    default=".",
    show_default=True,
)
@click.pass_context
def add_cmd(ctx: click.Context, final_dir: str) -> None:
    from chik.plotting.util import add_plot_directory

    try:
        add_plot_directory(ChikCliContext.set_default(ctx).root_path, final_dir)
        print(f"Successfully added: {final_dir}")
    except ValueError as e:
        print(e)


@plots_cmd.command("remove", help="Removes a directory of plots from config.yaml")
@click.option(
    "-d",
    "--final_dir",
    help="Final directory for plots (relative or absolute)",
    type=click.Path(),
    default=".",
    show_default=True,
)
@click.pass_context
def remove_cmd(ctx: click.Context, final_dir: str) -> None:
    from chik.plotting.util import remove_plot_directory

    remove_plot_directory(ChikCliContext.set_default(ctx).root_path, final_dir)


@plots_cmd.command("show", help="Shows the directory of current plots")
@click.pass_context
def show_cmd(ctx: click.Context) -> None:
    show_plots(ChikCliContext.set_default(ctx).root_path)



from __future__ import annotations

import contextlib
import random
from typing import Any

import pytest

from chik._tests.util.setup_nodes import SimulatorsAndWalletsServices
from chik._tests.util.time_out_assert import time_out_assert, time_out_assert_not_none
from chik.consensus.block_rewards import calculate_base_farmer_reward, calculate_pool_reward
from chik.rpc.full_node_rpc_client import FullNodeRpcClient
from chik.rpc.wallet_rpc_api import WalletRpcApi
from chik.rpc.wallet_rpc_client import WalletRpcClient
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.simulator.simulator_protocol import FarmNewBlockProtocol
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.peer_info import PeerInfo
from chik.util.bech32m import decode_puzzle_hash, encode_puzzle_hash
from chik.util.ints import uint16, uint32, uint64
from chik.wallet.did_wallet.did_wallet import DIDWallet
from chik.wallet.nft_wallet.nft_wallet import NFTWallet
from chik.wallet.nft_wallet.uncurry_nft import UncurriedNFT
from chik.wallet.util.address_type import AddressType
from chik.wallet.util.tx_config import DEFAULT_COIN_SELECTION_CONFIG, DEFAULT_TX_CONFIG


async def nft_count(wallet: NFTWallet) -> int:
    return await wallet.get_nft_count()


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_did(
    self_hostname: str, two_wallet_nodes: Any, trusted: Any, seeded_random: random.Random
) -> None:
    full_nodes, wallets, _ = two_wallet_nodes
    full_node_api: FullNodeSimulator = full_nodes[0]
    full_node_server = full_node_api.server
    wallet_node_0, server_0 = wallets[0]
    wallet_node_1, server_1 = wallets[1]
    wallet_0 = wallet_node_0.wallet_state_manager.main_wallet
    wallet_1 = wallet_node_1.wallet_state_manager.main_wallet
    ph_maker = await wallet_0.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_0.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_1.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_0.config["trusted_peers"] = {}
        wallet_node_1.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    # for _ in range(1, num_blocks):
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))
    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_0.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_0.get_confirmed_balance, funds)

    async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node_0.wallet_state_manager, wallet_0, uint64(1), action_scope
        )
    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)
    await time_out_assert(30, wallet_0.get_pending_change_balance, 0)

    hex_did_id = did_wallet.get_my_DID()
    did_id = bytes32.fromhex(hex_did_id)

    await time_out_assert(5, did_wallet.get_confirmed_balance, 1)

    nft_wallet_maker = await NFTWallet.create_new_nft_wallet(
        wallet_node_0.wallet_state_manager, wallet_0, name="NFT WALLET 1", did_id=did_id
    )

    nft_wallet_taker = await NFTWallet.create_new_nft_wallet(
        wallet_node_1.wallet_state_manager, wallet_1, name="NFT WALLET 2"
    )

    royalty_pc = uint16(300)
    royalty_addr = ph_maker

    mint_total = 10
    fee = uint64(100)
    metadata_list = [
        {
            "program": Program.to(
                [
                    ("u", ["https://www.chiknetwork.com/img/branding/chik-logo.svg"]),
                    ("h", bytes32.random(seeded_random).hex()),
                ]
            ),
            "royalty_pc": royalty_pc,
            "royalty_ph": royalty_addr,
        }
        for x in range(mint_total)
    ]

    target_list = [(await wallet_1.get_new_puzzlehash()) for x in range(mint_total)]

    async with nft_wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        await nft_wallet_maker.mint_from_did(
            metadata_list,
            action_scope,
            target_list=target_list,
            mint_number_start=1,
            mint_total=mint_total,
            fee=fee,
        )
    for record in action_scope.side_effects.transactions:
        if record.spend_bundle is not None:
            await time_out_assert_not_none(
                5, full_node_api.full_node.mempool_manager.get_spendbundle, record.spend_bundle.name()
            )
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    await time_out_assert(30, nft_count, mint_total, nft_wallet_taker)
    await time_out_assert(30, nft_count, 0, nft_wallet_maker)

    expected_xck_bal = funds - fee - mint_total - 1
    await time_out_assert(30, wallet_0.get_confirmed_balance, expected_xck_bal)

    nfts = await nft_wallet_taker.get_current_nfts()
    matched_data = dict(zip(target_list, metadata_list))

    # Check targets and metadata entries match in the final nfts
    for nft in nfts:
        mod, args = nft.full_puzzle.uncurry()
        unft = UncurriedNFT.uncurry(mod, args)
        assert isinstance(unft, UncurriedNFT)
        inner_args = unft.inner_puzzle.uncurry()[1]
        inner_ph = inner_args.at("rrrf").get_tree_hash()
        meta = unft.metadata.at("rfr").as_atom()
        # check that the target puzzle hashes of transferred nfts matches the metadata entry
        assert matched_data[inner_ph]["program"].at("rfr").as_atom() == meta
        # Check the did is set for each nft
        assert nft.minter_did == did_id


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_did_rpc(
    two_wallet_nodes_services: SimulatorsAndWalletsServices,
    trusted: Any,
    self_hostname: str,
    seeded_random: random.Random,
) -> None:
    [full_node_service], wallet_services, bt = two_wallet_nodes_services
    full_node_api: FullNodeSimulator = full_node_service._api
    full_node_server = full_node_api.server
    wallet_node_maker = wallet_services[0]._node
    wallet_node_taker = wallet_services[1]._node
    server_0 = wallet_node_maker.server
    server_1 = wallet_node_taker.server
    wallet_maker = wallet_node_maker.wallet_state_manager.main_wallet
    wallet_taker = wallet_node_taker.wallet_state_manager.main_wallet

    ph_maker = await wallet_maker.get_new_puzzlehash()
    ph_taker = await wallet_taker.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_maker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_taker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_maker.config["trusted_peers"] = {}
        wallet_node_taker.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_taker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_maker.get_confirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_confirmed_balance, funds)

    api_maker = WalletRpcApi(wallet_node_maker)
    api_taker = WalletRpcApi(wallet_node_taker)
    config = bt.config

    assert wallet_services[0].rpc_server is not None
    assert full_node_service.rpc_server is not None

    async with contextlib.AsyncExitStack() as exit_stack:
        client = await exit_stack.enter_async_context(
            WalletRpcClient.create_as_context(
                self_hostname,
                wallet_services[0].rpc_server.listen_port,
                wallet_services[0].root_path,
                wallet_services[0].config,
            )
        )
        client_node = await exit_stack.enter_async_context(
            FullNodeRpcClient.create_as_context(
                self_hostname,
                full_node_service.rpc_server.listen_port,
                full_node_service.root_path,
                full_node_service.config,
            )
        )

        async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_maker: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_maker.wallet_state_manager, wallet_maker, uint64(1), action_scope
            )
        await full_node_api.process_transaction_records(action_scope.side_effects.transactions)

        await time_out_assert(30, wallet_maker.get_pending_change_balance, 0)
        await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds - 1)
        await time_out_assert(30, wallet_maker.get_confirmed_balance, funds - 1)

        hex_did_id = did_wallet_maker.get_my_DID()
        hmr_did_id = encode_puzzle_hash(bytes32.from_hexstr(hex_did_id), AddressType.DID.hrp(config))

        await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)

        nft_wallet_maker = await api_maker.create_new_wallet(
            dict(wallet_type="nft_wallet", name="NFT WALLET 1", did_id=hmr_did_id)
        )
        assert isinstance(nft_wallet_maker, dict)
        assert nft_wallet_maker.get("success")

        nft_wallet_taker = await api_taker.create_new_wallet(dict(wallet_type="nft_wallet", name="NFT WALLET 2"))

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        n = 10
        metadata_list = [
            {
                "hash": bytes32.random(seeded_random).hex(),
                "uris": [f"https://data.com/{i}"],
                "meta_hash": bytes32.random(seeded_random).hex(),
                "meta_uris": [f"https://meatadata.com/{i}"],
                "license_hash": bytes32.random(seeded_random).hex(),
                "license_uris": [f"https://license.com/{i}"],
                "edition_number": i + 1,
                "edition_total": n,
            }
            for i in range(n)
        ]
        target_list = [encode_puzzle_hash((ph_taker), "xck") for x in range(n)]
        royalty_address = encode_puzzle_hash(bytes32.random(seeded_random), "xck")
        royalty_percentage = 300
        fee = 100
        required_amount = n + (fee * n)
        xck_coins = await client.select_coins(
            amount=required_amount, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG, wallet_id=wallet_maker.id()
        )
        funding_coin = xck_coins[0]
        assert funding_coin.amount >= required_amount
        funding_coin_dict = xck_coins[0].to_json_dict()
        chunk = 5
        next_coin = funding_coin
        did_coin = (
            await client.select_coins(amount=1, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG, wallet_id=2)
        )[0]
        did_lineage_parent = None
        spends = []
        nft_ids = set()
        for i in range(0, n, chunk):
            await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)
            resp = await client.nft_mint_bulk(
                wallet_id=nft_wallet_maker["wallet_id"],
                metadata_list=metadata_list[i : i + chunk],
                target_list=target_list[i : i + chunk],
                royalty_percentage=royalty_percentage,
                royalty_address=royalty_address,
                mint_number_start=i + 1,
                mint_total=n,
                xck_coins=[next_coin.to_json_dict()],
                xck_change_target=funding_coin_dict["puzzle_hash"],
                did_coin=did_coin.to_json_dict(),
                did_lineage_parent=did_lineage_parent,
                mint_from_did=True,
                fee=fee,
                tx_config=DEFAULT_TX_CONFIG,
            )
            sb = resp.spend_bundle
            did_lineage_parent = next(cn for cn in sb.removals() if cn.name() == did_coin.name()).parent_coin_info.hex()
            did_coin = next(
                cn for cn in sb.additions() if (cn.parent_coin_info == did_coin.name()) and (cn.amount == 1)
            )
            spends.append(sb)
            xck_adds = [c for c in sb.additions() if c.puzzle_hash == funding_coin.puzzle_hash]
            assert len(xck_adds) == 1
            next_coin = xck_adds[0]
            for nft_id in resp.nft_id_list:
                nft_ids.add(decode_puzzle_hash(nft_id))
        for sb in spends:
            push_resp = await client_node.push_tx(sb)
            assert push_resp["success"]
            await full_node_api.process_spend_bundles([sb])

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        async def get_taker_nfts() -> int:
            return int((await api_taker.nft_count_nfts({"wallet_id": nft_wallet_taker["wallet_id"]}))["count"])

        # We are using a long time out here because it can take a long time for the NFTs to show up
        # Even with only 10 NFTs it regularly takes longer than 30-40s for them to be found
        await time_out_assert(60, get_taker_nfts, n)

        # check NFT edition numbers
        nfts = (await api_taker.nft_get_nfts({"wallet_id": nft_wallet_taker["wallet_id"]}))["nft_list"]
        for nft in nfts:
            edition_num = nft.edition_number
            meta_dict = metadata_list[edition_num - 1]
            assert meta_dict["hash"] == nft.data_hash.hex()
            assert meta_dict["uris"] == nft.data_uris
            assert meta_dict["meta_hash"] == nft.metadata_hash.hex()
            assert meta_dict["meta_uris"] == nft.metadata_uris
            assert meta_dict["license_hash"] == nft.license_hash.hex()
            assert meta_dict["license_uris"] == nft.license_uris
            assert meta_dict["edition_number"] == nft.edition_number
            assert meta_dict["edition_total"] == nft.edition_total
            assert nft.launcher_id in nft_ids


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_did_rpc_no_royalties(
    two_wallet_nodes_services: SimulatorsAndWalletsServices,
    trusted: Any,
    self_hostname: str,
    seeded_random: random.Random,
) -> None:
    [full_node_service], wallet_services, bt = two_wallet_nodes_services
    full_node_api: FullNodeSimulator = full_node_service._api
    full_node_server = full_node_api.server
    wallet_node_maker = wallet_services[0]._node
    wallet_node_taker = wallet_services[1]._node
    server_0 = wallet_node_maker.server
    server_1 = wallet_node_taker.server
    wallet_maker = wallet_node_maker.wallet_state_manager.main_wallet
    wallet_taker = wallet_node_taker.wallet_state_manager.main_wallet

    ph_maker = await wallet_maker.get_new_puzzlehash()
    ph_taker = await wallet_taker.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_maker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_taker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_maker.config["trusted_peers"] = {}
        wallet_node_taker.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_taker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_maker.get_confirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_confirmed_balance, funds)

    api_maker = WalletRpcApi(wallet_node_maker)
    api_taker = WalletRpcApi(wallet_node_taker)
    config = bt.config

    assert wallet_services[0].rpc_server is not None
    assert full_node_service.rpc_server is not None

    async with contextlib.AsyncExitStack() as exit_stack:
        client = await exit_stack.enter_async_context(
            WalletRpcClient.create_as_context(
                self_hostname,
                wallet_services[0].rpc_server.listen_port,
                wallet_services[0].root_path,
                wallet_services[0].config,
            )
        )
        client_node = await exit_stack.enter_async_context(
            FullNodeRpcClient.create_as_context(
                self_hostname,
                full_node_service.rpc_server.listen_port,
                full_node_service.root_path,
                full_node_service.config,
            )
        )

        async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_maker: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_maker.wallet_state_manager, wallet_maker, uint64(1), action_scope
            )
        await full_node_api.process_transaction_records(action_scope.side_effects.transactions)

        await time_out_assert(30, wallet_maker.get_pending_change_balance, 0)
        await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds - 1)
        await time_out_assert(30, wallet_maker.get_confirmed_balance, funds - 1)

        hex_did_id = did_wallet_maker.get_my_DID()
        hmr_did_id = encode_puzzle_hash(bytes32.from_hexstr(hex_did_id), AddressType.DID.hrp(config))

        await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)

        nft_wallet_maker = await api_maker.create_new_wallet(
            dict(wallet_type="nft_wallet", name="NFT WALLET 1", did_id=hmr_did_id)
        )
        assert isinstance(nft_wallet_maker, dict)
        assert nft_wallet_maker.get("success")

        nft_wallet_taker = await api_taker.create_new_wallet(dict(wallet_type="nft_wallet", name="NFT WALLET 2"))

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        n = 10
        metadata_list = [
            {
                "hash": bytes32.random(seeded_random).hex(),
                "uris": [f"https://data.com/{i}"],
                "meta_hash": bytes32.random(seeded_random).hex(),
                "meta_uris": [f"https://meatadata.com/{i}"],
                "license_hash": bytes32.random(seeded_random).hex(),
                "license_uris": [f"https://license.com/{i}"],
                "edition_number": i + 1,
                "edition_total": n,
            }
            for i in range(n)
        ]
        target_list = [encode_puzzle_hash((ph_taker), "xck") for x in range(n)]
        royalty_address = None
        royalty_percentage = None
        required_amount = n
        xck_coins = await client.select_coins(
            amount=required_amount, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG, wallet_id=wallet_maker.id()
        )
        funding_coin = xck_coins[0]
        assert funding_coin.amount >= required_amount
        funding_coin_dict = xck_coins[0].to_json_dict()
        chunk = 5
        next_coin = funding_coin
        did_coin = (
            await client.select_coins(amount=1, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG, wallet_id=2)
        )[0]
        did_lineage_parent = None
        spends = []

        for i in range(0, n, chunk):
            await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)
            resp = await client.nft_mint_bulk(
                wallet_id=nft_wallet_maker["wallet_id"],
                metadata_list=metadata_list[i : i + chunk],
                target_list=target_list[i : i + chunk],
                royalty_percentage=royalty_percentage,
                royalty_address=royalty_address,
                mint_number_start=i + 1,
                mint_total=n,
                xck_coins=[next_coin.to_json_dict()],
                xck_change_target=funding_coin_dict["puzzle_hash"],
                did_coin=did_coin.to_json_dict(),
                did_lineage_parent=did_lineage_parent,
                mint_from_did=True,
                tx_config=DEFAULT_TX_CONFIG,
            )
            sb = resp.spend_bundle
            did_lineage_parent = next(cn for cn in sb.removals() if cn.name() == did_coin.name()).parent_coin_info.hex()
            did_coin = next(
                cn for cn in sb.additions() if (cn.parent_coin_info == did_coin.name()) and (cn.amount == 1)
            )
            spends.append(sb)
            xck_adds = [c for c in sb.additions() if c.puzzle_hash == funding_coin.puzzle_hash]
            assert len(xck_adds) == 1
            next_coin = xck_adds[0]

        for sb in spends:
            push_resp = await client_node.push_tx(sb)
            assert push_resp["success"]
            await full_node_api.process_spend_bundles([sb])

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        async def get_taker_nfts() -> int:
            return int((await api_taker.nft_count_nfts({"wallet_id": nft_wallet_taker["wallet_id"]}))["count"])

        await time_out_assert(60, get_taker_nfts, n)


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_did_multiple_xck(
    self_hostname: str, two_wallet_nodes: Any, trusted: Any, seeded_random: random.Random
) -> None:
    full_nodes, wallets, _ = two_wallet_nodes
    full_node_api: FullNodeSimulator = full_nodes[0]
    full_node_server = full_node_api.server
    wallet_node_0, server_0 = wallets[0]
    wallet_node_1, server_1 = wallets[1]
    wallet_maker = wallet_node_0.wallet_state_manager.main_wallet
    wallet_taker = wallet_node_1.wallet_state_manager.main_wallet
    ph_maker = await wallet_maker.get_new_puzzlehash()
    ph_taker = await wallet_taker.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_0.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_1.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_0.config["trusted_peers"] = {}
        wallet_node_1.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_maker.get_confirmed_balance, funds)

    async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node_0.wallet_state_manager, wallet_maker, uint64(1), action_scope
        )
    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)
    await time_out_assert(30, wallet_maker.get_pending_change_balance, 0)

    hex_did_id = did_wallet.get_my_DID()
    did_id = bytes32.fromhex(hex_did_id)

    await time_out_assert(5, did_wallet.get_confirmed_balance, 1)

    nft_wallet_maker = await NFTWallet.create_new_nft_wallet(
        wallet_node_0.wallet_state_manager, wallet_maker, name="NFT WALLET 1", did_id=did_id
    )

    nft_wallet_taker = await NFTWallet.create_new_nft_wallet(
        wallet_node_1.wallet_state_manager, wallet_taker, name="NFT WALLET 2"
    )

    # construct sample metadata
    metadata = Program.to(
        [
            ("u", ["https://www.chiknetwork.com/img/branding/chik-logo.svg"]),
            ("h", "0xD4584AD463139FA8C0D9F68F4B59F185"),
        ]
    )
    royalty_pc = uint16(300)
    royalty_addr = ph_maker

    mint_total = 1
    fee = uint64(100)
    metadata_list = [
        {"program": metadata, "royalty_pc": royalty_pc, "royalty_ph": royalty_addr} for x in range(mint_total)
    ]

    # Grab two coins for testing that we can create a bulk minting with more than 1 xck coin
    async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
        xck_coins_1 = await wallet_maker.select_coins(amount=10000, action_scope=action_scope)
        xck_coins_2 = await wallet_maker.select_coins(
            amount=10000,
            action_scope=action_scope,
        )
    xck_coins = xck_coins_1.union(xck_coins_2)

    target_list = [ph_taker for x in range(mint_total)]

    async with nft_wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        await nft_wallet_maker.mint_from_did(
            metadata_list,
            action_scope,
            target_list=target_list,
            mint_number_start=1,
            mint_total=mint_total,
            xck_coins=xck_coins,
            fee=fee,
        )
    sb = action_scope.side_effects.transactions[0].spend_bundle
    assert sb is not None

    await time_out_assert_not_none(5, full_node_api.full_node.mempool_manager.get_spendbundle, sb.name())
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    await time_out_assert(30, nft_count, mint_total, nft_wallet_taker)
    await time_out_assert(30, nft_count, 0, nft_wallet_maker)

    # confirm that the spend uses the right amount of xck
    expected_xck_bal = funds - fee - mint_total - 1
    await time_out_assert(30, wallet_maker.get_confirmed_balance, expected_xck_bal)


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_xck(
    self_hostname: str, two_wallet_nodes: Any, trusted: Any, seeded_random: random.Random
) -> None:
    full_nodes, wallets, _ = two_wallet_nodes
    full_node_api: FullNodeSimulator = full_nodes[0]
    full_node_server = full_node_api.server
    wallet_node_0, server_0 = wallets[0]
    wallet_node_1, server_1 = wallets[1]
    wallet_0 = wallet_node_0.wallet_state_manager.main_wallet
    wallet_1 = wallet_node_1.wallet_state_manager.main_wallet
    ph_maker = await wallet_0.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_0.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_1.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_0.config["trusted_peers"] = {}
        wallet_node_1.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    # for _ in range(1, num_blocks):
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))
    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_0.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_0.get_confirmed_balance, funds)

    async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node_0.wallet_state_manager, wallet_0, uint64(1), action_scope
        )
    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)
    await time_out_assert(30, wallet_0.get_pending_change_balance, 0)

    hex_did_id = did_wallet.get_my_DID()
    did_id = bytes32.fromhex(hex_did_id)

    await time_out_assert(5, did_wallet.get_confirmed_balance, 1)

    nft_wallet_maker = await NFTWallet.create_new_nft_wallet(
        wallet_node_0.wallet_state_manager, wallet_0, name="NFT WALLET 1", did_id=did_id
    )

    nft_wallet_taker = await NFTWallet.create_new_nft_wallet(
        wallet_node_1.wallet_state_manager, wallet_1, name="NFT WALLET 2"
    )

    royalty_pc = uint16(300)
    royalty_addr = ph_maker

    mint_total = 1
    fee = uint64(100)
    metadata_list = [
        {
            "program": Program.to(
                [
                    ("u", ["https://www.chiknetwork.com/img/branding/chik-logo.svg"]),
                    ("h", bytes32.random(seeded_random).hex()),
                ]
            ),
            "royalty_pc": royalty_pc,
            "royalty_ph": royalty_addr,
        }
        for x in range(mint_total)
    ]

    target_list = [(await wallet_1.get_new_puzzlehash()) for x in range(mint_total)]

    async with nft_wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        await nft_wallet_maker.mint_from_xck(
            metadata_list,
            action_scope,
            target_list=target_list,
            mint_number_start=1,
            mint_total=mint_total,
            fee=fee,
        )

    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)

    await time_out_assert(30, nft_count, mint_total, nft_wallet_taker)
    await time_out_assert(30, nft_count, 0, nft_wallet_maker)

    expected_xck_bal = funds - fee - mint_total - 1
    await time_out_assert(30, wallet_0.get_confirmed_balance, expected_xck_bal)

    nfts = await nft_wallet_taker.get_current_nfts()
    matched_data = dict(zip(target_list, metadata_list))

    # Check targets and metadata entries match in the final nfts
    for nft in nfts:
        mod, args = nft.full_puzzle.uncurry()
        unft = UncurriedNFT.uncurry(mod, args)
        assert isinstance(unft, UncurriedNFT)
        inner_args = unft.inner_puzzle.uncurry()[1]
        inner_ph = inner_args.at("rrrf").get_tree_hash()
        meta = unft.metadata.at("rfr").as_atom()
        # check that the target puzzle hashes of transferred nfts matches the metadata entry
        assert matched_data[inner_ph]["program"].at("rfr").as_atom() == meta
        assert not nft.minter_did


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_xck_rpc(
    two_wallet_nodes_services: SimulatorsAndWalletsServices,
    trusted: Any,
    self_hostname: str,
    seeded_random: random.Random,
) -> None:
    [full_node_service], wallet_services, bt = two_wallet_nodes_services
    full_node_api: FullNodeSimulator = full_node_service._api
    full_node_server = full_node_api.server
    wallet_node_maker = wallet_services[0]._node
    wallet_node_taker = wallet_services[1]._node
    server_0 = wallet_node_maker.server
    server_1 = wallet_node_taker.server
    wallet_maker = wallet_node_maker.wallet_state_manager.main_wallet
    wallet_taker = wallet_node_taker.wallet_state_manager.main_wallet

    ph_maker = await wallet_maker.get_new_puzzlehash()
    ph_taker = await wallet_taker.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_maker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_taker.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_maker.config["trusted_peers"] = {}
        wallet_node_taker.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_taker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_maker.get_confirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_taker.get_confirmed_balance, funds)

    api_maker = WalletRpcApi(wallet_node_maker)
    api_taker = WalletRpcApi(wallet_node_taker)
    config = bt.config

    assert wallet_services[0].rpc_server is not None
    assert full_node_service.rpc_server is not None

    async with contextlib.AsyncExitStack() as exit_stack:
        client = await exit_stack.enter_async_context(
            WalletRpcClient.create_as_context(
                self_hostname,
                wallet_services[0].rpc_server.listen_port,
                wallet_services[0].root_path,
                wallet_services[0].config,
            )
        )
        client_node = await exit_stack.enter_async_context(
            FullNodeRpcClient.create_as_context(
                self_hostname,
                full_node_service.rpc_server.listen_port,
                full_node_service.root_path,
                full_node_service.config,
            )
        )

        async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_maker: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_maker.wallet_state_manager, wallet_maker, uint64(1), action_scope
            )
        await full_node_api.process_transaction_records(action_scope.side_effects.transactions)

        await time_out_assert(30, wallet_maker.get_pending_change_balance, 0)
        await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds - 1)
        await time_out_assert(30, wallet_maker.get_confirmed_balance, funds - 1)

        hex_did_id = did_wallet_maker.get_my_DID()
        hmr_did_id = encode_puzzle_hash(bytes32.from_hexstr(hex_did_id), AddressType.DID.hrp(config))

        await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)

        nft_wallet_maker = await api_maker.create_new_wallet(
            dict(wallet_type="nft_wallet", name="NFT WALLET 1", did_id=hmr_did_id)
        )
        assert isinstance(nft_wallet_maker, dict)
        assert nft_wallet_maker.get("success")

        nft_wallet_taker = await api_taker.create_new_wallet(dict(wallet_type="nft_wallet", name="NFT WALLET 2"))

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        n = 10
        metadata_list = [
            {
                "hash": bytes32.random(seeded_random).hex(),
                "uris": [f"https://data.com/{i}"],
                "meta_hash": bytes32.random(seeded_random).hex(),
                "meta_uris": [f"https://meatadata.com/{i}"],
                "license_hash": bytes32.random(seeded_random).hex(),
                "license_uris": [f"https://license.com/{i}"],
                "edition_number": i + 1,
                "edition_total": n,
            }
            for i in range(n)
        ]
        target_list = [encode_puzzle_hash((ph_taker), "xck") for x in range(n)]
        royalty_address = encode_puzzle_hash(bytes32.random(seeded_random), "xck")
        royalty_percentage = 300
        fee = 100
        required_amount = n + (fee * n)
        xck_coins = await client.select_coins(
            amount=required_amount, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG, wallet_id=wallet_maker.id()
        )
        funding_coin = xck_coins[0]
        assert funding_coin.amount >= required_amount
        funding_coin_dict = xck_coins[0].to_json_dict()
        chunk = 5
        next_coin = funding_coin
        spends = []

        for i in range(0, n, chunk):
            await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_maker, timeout=20)
            resp = await client.nft_mint_bulk(
                wallet_id=nft_wallet_maker["wallet_id"],
                metadata_list=metadata_list[i : i + chunk],
                target_list=target_list[i : i + chunk],
                royalty_percentage=royalty_percentage,
                royalty_address=royalty_address,
                mint_number_start=i + 1,
                mint_total=n,
                xck_coins=[next_coin.to_json_dict()],
                xck_change_target=funding_coin_dict["puzzle_hash"],
                mint_from_did=False,
                fee=fee,
                tx_config=DEFAULT_TX_CONFIG,
            )
            sb = resp.spend_bundle
            spends.append(sb)
            xck_adds = [c for c in sb.additions() if c.puzzle_hash == funding_coin.puzzle_hash]
            assert len(xck_adds) == 1
            next_coin = xck_adds[0]

        for sb in spends:
            push_resp = await client_node.push_tx(sb)
            assert push_resp["success"]
            await full_node_api.process_spend_bundles([sb])

        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

        async def get_taker_nfts() -> int:
            return int((await api_taker.nft_count_nfts({"wallet_id": nft_wallet_taker["wallet_id"]}))["count"])

        # We are using a long time out here because it can take a long time for the NFTs to show up
        # Even with only 10 NFTs it regularly takes longer than 30-40s for them to be found
        await time_out_assert(60, get_taker_nfts, n)

        # check NFT edition numbers
        nfts = (await api_taker.nft_get_nfts({"wallet_id": nft_wallet_taker["wallet_id"]}))["nft_list"]
        for nft in nfts:
            edition_num = nft.edition_number
            meta_dict = metadata_list[edition_num - 1]
            assert meta_dict["hash"] == nft.data_hash.hex()
            assert meta_dict["uris"] == nft.data_uris
            assert meta_dict["meta_hash"] == nft.metadata_hash.hex()
            assert meta_dict["meta_uris"] == nft.metadata_uris
            assert meta_dict["license_hash"] == nft.license_hash.hex()
            assert meta_dict["license_uris"] == nft.license_uris
            assert meta_dict["edition_number"] == nft.edition_number
            assert meta_dict["edition_total"] == nft.edition_total


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_nft_mint_from_xck_multiple_xck(
    self_hostname: str, two_wallet_nodes: Any, trusted: Any, seeded_random: random.Random
) -> None:
    full_nodes, wallets, _ = two_wallet_nodes
    full_node_api: FullNodeSimulator = full_nodes[0]
    full_node_server = full_node_api.server
    wallet_node_0, server_0 = wallets[0]
    wallet_node_1, server_1 = wallets[1]
    wallet_maker = wallet_node_0.wallet_state_manager.main_wallet
    wallet_taker = wallet_node_1.wallet_state_manager.main_wallet
    ph_maker = await wallet_maker.get_new_puzzlehash()
    ph_taker = await wallet_taker.get_new_puzzlehash()
    ph_token = bytes32.random(seeded_random)

    if trusted:
        wallet_node_0.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
        wallet_node_1.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_0.config["trusted_peers"] = {}
        wallet_node_1.config["trusted_peers"] = {}

    await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_maker))
    await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(ph_token))

    funds = calculate_pool_reward(uint32(1)) + calculate_base_farmer_reward(uint32(1))

    await time_out_assert(30, wallet_maker.get_unconfirmed_balance, funds)
    await time_out_assert(30, wallet_maker.get_confirmed_balance, funds)

    async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node_0.wallet_state_manager, wallet_maker, uint64(1), action_scope
        )
    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)
    await time_out_assert(30, wallet_maker.get_pending_change_balance, 0)

    hex_did_id = did_wallet.get_my_DID()
    did_id = bytes32.fromhex(hex_did_id)

    await time_out_assert(5, did_wallet.get_confirmed_balance, 1)

    nft_wallet_maker = await NFTWallet.create_new_nft_wallet(
        wallet_node_0.wallet_state_manager, wallet_maker, name="NFT WALLET 1", did_id=did_id
    )

    nft_wallet_taker = await NFTWallet.create_new_nft_wallet(
        wallet_node_1.wallet_state_manager, wallet_taker, name="NFT WALLET 2"
    )

    # construct sample metadata
    metadata = Program.to(
        [
            ("u", ["https://www.chiknetwork.com/img/branding/chik-logo.svg"]),
            ("h", "0xD4584AD463139FA8C0D9F68F4B59F185"),
        ]
    )
    royalty_pc = uint16(300)
    royalty_addr = ph_maker

    mint_total = 1
    fee = uint64(100)
    metadata_list = [
        {"program": metadata, "royalty_pc": royalty_pc, "royalty_ph": royalty_addr} for x in range(mint_total)
    ]

    # Grab two coins for testing that we can create a bulk minting with more than 1 xck coin
    async with wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
        xck_coins_1 = await wallet_maker.select_coins(amount=10000, action_scope=action_scope)
        xck_coins_2 = await wallet_maker.select_coins(
            amount=10000,
            action_scope=action_scope,
        )
    xck_coins = xck_coins_1.union(xck_coins_2)

    target_list = [ph_taker for x in range(mint_total)]

    async with nft_wallet_maker.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        await nft_wallet_maker.mint_from_xck(
            metadata_list,
            action_scope,
            target_list=target_list,
            mint_number_start=1,
            mint_total=mint_total,
            xck_coins=xck_coins,
            fee=fee,
        )

    await full_node_api.process_transaction_records(action_scope.side_effects.transactions)

    await time_out_assert(30, nft_count, mint_total, nft_wallet_taker)
    await time_out_assert(30, nft_count, 0, nft_wallet_maker)

    # confirm that the spend uses the right amount of xck
    expected_xck_bal = funds - fee - mint_total - 1
    await time_out_assert(30, wallet_maker.get_confirmed_balance, expected_xck_bal)



from __future__ import annotations

from typing import Any

from chik.plot_sync.util import ErrorCodes, State
from chik.protocols.harvester_protocol import PlotSyncIdentifier
from chik.server.outbound_message import NodeType
from chik.util.ints import uint64


class PlotSyncException(Exception):
    def __init__(self, message: str, error_code: ErrorCodes) -> None:
        super().__init__(message)
        self.error_code = error_code


class AlreadyStartedError(Exception):
    def __init__(self) -> None:
        super().__init__("Already started!")


class InvalidValueError(PlotSyncException):
    def __init__(self, message: str, actual: Any, expected: Any, error_code: ErrorCodes) -> None:
        super().__init__(f"{message}: Actual {actual}, Expected {expected}", error_code)


class InvalidIdentifierError(InvalidValueError):
    def __init__(self, actual_identifier: PlotSyncIdentifier, expected_identifier: PlotSyncIdentifier) -> None:
        super().__init__("Invalid identifier", actual_identifier, expected_identifier, ErrorCodes.invalid_identifier)
        self.actual_identifier: PlotSyncIdentifier = actual_identifier
        self.expected_identifier: PlotSyncIdentifier = expected_identifier


class InvalidLastSyncIdError(InvalidValueError):
    def __init__(self, actual: uint64, expected: uint64) -> None:
        super().__init__("Invalid last-sync-id", actual, expected, ErrorCodes.invalid_last_sync_id)


class InvalidConnectionTypeError(InvalidValueError):
    def __init__(self, actual: NodeType, expected: NodeType) -> None:
        super().__init__("Unexpected connection type", actual, expected, ErrorCodes.invalid_connection_type)


class PlotAlreadyAvailableError(PlotSyncException):
    def __init__(self, state: State, path: str) -> None:
        super().__init__(f"{state.name}: Plot already available - {path}", ErrorCodes.plot_already_available)


class PlotNotAvailableError(PlotSyncException):
    def __init__(self, state: State, path: str) -> None:
        super().__init__(f"{state.name}: Plot not available - {path}", ErrorCodes.plot_not_available)


class SyncIdsMatchError(PlotSyncException):
    def __init__(self, state: State, sync_id: uint64) -> None:
        super().__init__(f"{state.name}: Sync ids are equal - {sync_id}", ErrorCodes.sync_ids_match)



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path

import aiohttp
import anyio
import click
import pytest

# TODO: update after resolution in https://github.com/pytest-dev/pytest/issues/7469
from _pytest.capture import CaptureFixture

import chik._tests
from chik._tests.util.misc import Marks, datacases
from chik.cmds.gh import Per, TestCMD, get_gh_token

test_root = Path(chik._tests.__file__).parent


@dataclass
class InvalidOnlyCase:
    only: Path
    per: Per
    exists: bool
    marks: Marks = ()

    @property
    def id(self) -> str:
        return f"{self.per}: {self.only}"


@datacases(
    InvalidOnlyCase(only=Path("does_not_exist.py"), per="directory", exists=False),
    InvalidOnlyCase(only=Path("pools/test_pool_rpc.py"), per="directory", exists=True),
    InvalidOnlyCase(only=Path("does_not_exist/"), per="file", exists=False),
    InvalidOnlyCase(only=Path("pools/"), per="file", exists=True),
)
@pytest.mark.anyio
async def test_invalid_only(case: InvalidOnlyCase) -> None:
    cmd = TestCMD(only=case.only, per=case.per)

    if case.exists:
        assert test_root.joinpath(case.only).exists()
        explanation = "wrong type"
        if case.per == "directory":
            assert test_root.joinpath(case.only).is_file()
        else:
            assert test_root.joinpath(case.only).is_dir()
    else:
        assert not test_root.joinpath(case.only).exists()
        explanation = "does not exist"

    with pytest.raises(click.ClickException, match=rf"\bto be a {re.escape(case.per)}\b.*\b{re.escape(explanation)}\b"):
        await cmd.run()


@pytest.mark.skip("considering inclusion, but not yet")
@pytest.mark.anyio
async def test_successfully_dispatches(
    capsys: CaptureFixture[str],
) -> None:
    cmd = TestCMD(
        # TODO: stop hardcoding here
        owner="chik-network",
        repository="chik-blockchain",
        per="file",
        only=Path("util/test_errors.py"),
        duplicates=2,
        oses=["linux", "macos-arm"],
        full_python_matrix=True,
        open_browser=False,
    )

    capsys.readouterr()
    await cmd.run()
    stdout, stderr = capsys.readouterr()

    assert len(stderr.strip()) == 0
    for line in stdout.splitlines():
        match = re.search(r"(?<=\brun api url: )(?P<url>.*)", line)
        if match is None:
            continue
        url = match.group("url")
        break
    else:
        pytest.fail(f"Failed to find run url in: {stdout}")

    token = await get_gh_token()
    headers = {"Authorization": f"Bearer {token}"}

    async with aiohttp.ClientSession(raise_for_status=True, headers=headers) as client:
        while True:
            async with client.get(url) as response:
                d = await response.json()
                jobs_url = d["jobs_url"]
                conclusion = d["conclusion"]

                print("conclusion:", conclusion)
                if conclusion is None:
                    await anyio.sleep(5)
                    continue

                break

        async with client.get(jobs_url) as response:
            d = await response.json()
            jobs = d["jobs"]

        by_name = {job["name"]: job for job in jobs}

        assert by_name["Configure matrix"]["conclusion"] == "success"
        assert by_name["macos-intel"]["conclusion"] == "skipped"
        assert by_name["windows"]["conclusion"] == "skipped"

        versions = ["3.9", "3.10", "3.11", "3.12"]
        runs_by_name: dict[str, list[str]] = {name: [] for name in ["ubuntu", "macos-arm"]}
        for name in by_name:
            platform, _, rest = name.partition(" / ")

            jobs = runs_by_name.get(platform)
            if jobs is None:
                continue

            jobs.append(rest)

        expected = len(versions) * cmd.duplicates
        print("expected:", expected)
        print("runs_by_name:", runs_by_name)
        assert len({expected, *(len(runs) for runs in runs_by_name.values())}) == 1



from __future__ import annotations

import functools
from collections.abc import Iterable
from dataclasses import dataclass, replace
from enum import IntEnum
from typing import Optional, TypeVar

from klvm.casts import int_to_bytes

from chik.types.blockchain_format.coin import Coin, coin_as_list
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, make_spend
from chik.util.hash import std_hash
from chik.util.ints import uint16, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.cat_wallet.cat_utils import CAT_MOD, construct_cat_puzzle
from chik.wallet.conditions import AssertCoinAnnouncement
from chik.wallet.lineage_proof import LineageProof, LineageProofField
from chik.wallet.payment import Payment
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.puzzles.singleton_top_layer_v1_1 import SINGLETON_LAUNCHER_HASH, SINGLETON_MOD_HASH
from chik.wallet.uncurried_puzzle import UncurriedPuzzle, uncurry_puzzle
from chik.wallet.util.curry_and_treehash import curry_and_treehash
from chik.wallet.vc_wallet.vc_drivers import (
    COVENANT_LAYER_HASH,
    EML_TP_COVENANT_ADAPTER_HASH,
    EXTIGENT_METADATA_LAYER_HASH,
    GUARANTEED_NIL_TP_HASH,
    P2_ANNOUNCED_DELEGATED_PUZZLE_HASH,
    create_did_tp,
    create_eml_covenant_morpher,
)

# Mods

CREDENTIAL_RESTRICTION: Program = load_klvm_maybe_recompile(
    "credential_restriction.clsp",
    package_or_requirement="chik.wallet.vc_wallet.cr_puzzles",
    include_standard_libraries=True,
)
CREDENTIAL_RESTRICTION_HASH: bytes32 = CREDENTIAL_RESTRICTION.get_tree_hash()
PROOF_FLAGS_CHECKER: Program = load_klvm_maybe_recompile(
    "flag_proofs_checker.clsp",
    package_or_requirement="chik.wallet.vc_wallet.cr_puzzles",
    include_standard_libraries=True,
)
PENDING_VC_ANNOUNCEMENT: Program = load_klvm_maybe_recompile(
    "conditions_w_fee_announce.clsp",
    package_or_requirement="chik.wallet.vc_wallet.cr_puzzles",
    include_standard_libraries=True,
)
CREDENTIAL_STRUCT: Program = Program.to(
    (
        (
            (
                SINGLETON_MOD_HASH,
                SINGLETON_LAUNCHER_HASH,
            ),
            (
                EXTIGENT_METADATA_LAYER_HASH,
                EML_TP_COVENANT_ADAPTER_HASH,
            ),
        ),
        (
            curry_and_treehash(
                Program.to((1, EXTIGENT_METADATA_LAYER_HASH)).get_tree_hash_precalc(EXTIGENT_METADATA_LAYER_HASH),
                Program.to(EXTIGENT_METADATA_LAYER_HASH).get_tree_hash(),
                Program.to(None).get_tree_hash(),
                GUARANTEED_NIL_TP_HASH,
                Program.to(GUARANTEED_NIL_TP_HASH).get_tree_hash(),
                P2_ANNOUNCED_DELEGATED_PUZZLE_HASH,
            ),
            (
                Program.to(
                    int_to_bytes(2) + Program.to((1, COVENANT_LAYER_HASH)).get_tree_hash_precalc(COVENANT_LAYER_HASH)
                ),
                Program.to(
                    (
                        [
                            4,
                            (1, create_eml_covenant_morpher(create_did_tp().get_tree_hash())),
                            [4, (1, create_did_tp()), 1],
                        ],
                        None,
                    )
                ).get_tree_hash(),
            ),
        ),
    ),
)
CREDENTIAL_STRUCT_HASH: bytes32 = CREDENTIAL_STRUCT.get_tree_hash()


# Basic drivers
def construct_cr_layer(
    authorized_providers: list[bytes32],
    proofs_checker: Program,
    inner_puzzle: Program,
) -> Program:
    first_curry: Program = CREDENTIAL_RESTRICTION.curry(
        CREDENTIAL_STRUCT,
        authorized_providers,
        proofs_checker,
    )
    return first_curry.curry(first_curry.get_tree_hash(), inner_puzzle)


def construct_cr_layer_hash(
    authorized_providers_hash: bytes32,
    proofs_checker_hash: bytes32,
    inner_puzzle_hash: bytes32,
) -> bytes32:
    first_curry_hash: bytes32 = curry_and_treehash(
        Program.to((1, CREDENTIAL_RESTRICTION_HASH)).get_tree_hash_precalc(CREDENTIAL_RESTRICTION_HASH),
        CREDENTIAL_STRUCT_HASH,
        authorized_providers_hash,
        proofs_checker_hash,
    )
    first_curry_hash_hash: bytes32 = Program.to(first_curry_hash).get_tree_hash()
    final_hash: bytes32 = curry_and_treehash(
        Program.to((1, first_curry_hash)).get_tree_hash_precalc(first_curry_hash),
        first_curry_hash_hash,
        inner_puzzle_hash,
    )
    return final_hash


def match_cr_layer(
    uncurried_puzzle: UncurriedPuzzle,
) -> Optional[tuple[list[bytes32], Program, Program]]:
    extra_uncurried_puzzle = uncurry_puzzle(uncurried_puzzle.mod)
    if extra_uncurried_puzzle.mod == CREDENTIAL_RESTRICTION:
        return (
            [bytes32(provider.as_atom()) for provider in extra_uncurried_puzzle.args.at("rf").as_iter()],
            extra_uncurried_puzzle.args.at("rrf"),
            uncurried_puzzle.args.at("rf"),
        )
    else:
        return None


def solve_cr_layer(
    proof_of_inclusions: Program,
    proof_checker_solution: Program,
    provider_id: bytes32,
    vc_launcher_id: bytes32,
    vc_inner_puzhash: bytes32,
    my_coin_id: bytes32,
    inner_solution: Program,
) -> Program:
    solution: Program = Program.to(
        [
            proof_of_inclusions,
            proof_checker_solution,
            provider_id,
            vc_launcher_id,
            vc_inner_puzhash,
            my_coin_id,
            inner_solution,
        ]
    )
    return solution


# For the "pending approval" state
def construct_pending_approval_state(puzzle_hash: bytes32, amount: uint64) -> Program:
    return PENDING_VC_ANNOUNCEMENT.curry(Program.to([[51, puzzle_hash, amount, [puzzle_hash]]]))


_T_CRCAT = TypeVar("_T_CRCAT", bound="CRCAT")


@dataclass(frozen=True)
class CRCAT:
    coin: Coin
    tail_hash: bytes32
    lineage_proof: LineageProof
    authorized_providers: list[bytes32]
    proofs_checker: Program
    inner_puzzle_hash: bytes32

    @classmethod
    def launch(
        cls: type[_T_CRCAT],
        # General CAT launching info
        origin_coin: Coin,
        payment: Payment,
        tail: Program,
        tail_solution: Program,
        # CR Layer params
        authorized_providers: list[bytes32],
        proofs_checker: Program,
        # Probably never need this but some tail might
        optional_lineage_proof: Optional[LineageProof] = None,
    ) -> tuple[Program, CoinSpend, CRCAT]:
        """
        Launch a new CR-CAT from XCK.

        Returns a delegated puzzle to run that creates the eve CAT, an eve coin spend of the CAT, and the expected class
        representation after all relevant coin spends have been confirmed on chain.
        """
        tail_hash: bytes32 = tail.get_tree_hash()

        new_cr_layer_hash: bytes32 = construct_cr_layer(
            authorized_providers,
            proofs_checker,
            payment.puzzle_hash,  # type: ignore
        ).get_tree_hash_precalc(payment.puzzle_hash)
        new_cat_puzhash = construct_cat_puzzle(CAT_MOD, tail_hash, new_cr_layer_hash).get_tree_hash_precalc(
            new_cr_layer_hash
        )

        eve_innerpuz: Program = Program.to(
            (
                1,
                [
                    [51, new_cr_layer_hash, payment.amount, payment.memos],
                    [51, None, -113, tail, tail_solution],
                    [60, None],
                    [1, payment.puzzle_hash, authorized_providers, proofs_checker],
                ],
            )
        )
        eve_cat_puzzle: Program = construct_cat_puzzle(
            CAT_MOD,
            tail_hash,
            eve_innerpuz,
        )
        eve_cat_puzzle_hash: bytes32 = eve_cat_puzzle.get_tree_hash()

        eve_coin: Coin = Coin(origin_coin.name(), eve_cat_puzzle_hash, payment.amount)
        dpuz: Program = Program.to(
            (
                1,
                [
                    [51, eve_cat_puzzle_hash, payment.amount],
                    [61, std_hash(eve_coin.name())],
                ],
            )
        )

        eve_proof: LineageProof = LineageProof(
            eve_coin.parent_coin_info,
            eve_innerpuz.get_tree_hash(),
            uint64(eve_coin.amount),
        )

        return (
            dpuz,
            make_spend(
                eve_coin,
                eve_cat_puzzle,
                Program.to(  # solve_cat
                    [
                        None,
                        optional_lineage_proof,
                        eve_coin.name(),
                        coin_as_list(eve_coin),
                        eve_proof.to_program(),
                        0,
                        0,
                    ]
                ),
            ),
            CRCAT(
                Coin(eve_coin.name(), new_cat_puzhash, payment.amount),
                tail_hash,
                eve_proof,
                authorized_providers,
                proofs_checker,
                payment.puzzle_hash,
            ),
        )

    def construct_puzzle(self, inner_puzzle: Program) -> Program:
        return construct_cat_puzzle(
            CAT_MOD,
            self.tail_hash,
            self.construct_cr_layer(inner_puzzle),
        )

    def construct_cr_layer(self, inner_puzzle: Program) -> Program:
        return construct_cr_layer(
            self.authorized_providers,
            self.proofs_checker,
            inner_puzzle,
        )

    @staticmethod
    def is_cr_cat(puzzle_reveal: UncurriedPuzzle) -> tuple[bool, str]:
        """
        This takes an (uncurried) puzzle reveal and returns a boolean for whether the puzzle is a CR-CAT and an error
        message for if the puzzle is a mismatch.
        """
        if puzzle_reveal.mod != CAT_MOD:
            return False, "top most layer is not a CAT"  # pragma: no cover
        layer_below_cat: UncurriedPuzzle = uncurry_puzzle(uncurry_puzzle(puzzle_reveal.args.at("rrf")).mod)
        if layer_below_cat.mod != CREDENTIAL_RESTRICTION:
            return False, "CAT is not credential restricted"  # pragma: no cover

        return True, ""

    @staticmethod
    def get_inner_puzzle(puzzle_reveal: UncurriedPuzzle) -> Program:  # pragma: no cover
        return uncurry_puzzle(puzzle_reveal.args.at("rrf")).args.at("rf")

    @staticmethod
    def get_inner_solution(solution: Program) -> Program:  # pragma: no cover
        return solution.at("f").at("rrrrrrf")

    @classmethod
    def get_current_from_coin_spend(cls: type[_T_CRCAT], spend: CoinSpend) -> CRCAT:  # pragma: no cover
        uncurried_puzzle: UncurriedPuzzle = uncurry_puzzle(spend.puzzle_reveal)
        first_uncurried_cr_layer: UncurriedPuzzle = uncurry_puzzle(uncurried_puzzle.args.at("rrf"))
        second_uncurried_cr_layer: UncurriedPuzzle = uncurry_puzzle(first_uncurried_cr_layer.mod)
        lineage_proof = LineageProof.from_program(
            spend.solution.to_program().at("rf"),
            [LineageProofField.PARENT_NAME, LineageProofField.INNER_PUZZLE_HASH, LineageProofField.AMOUNT],
        )
        return CRCAT(
            spend.coin,
            bytes32(uncurried_puzzle.args.at("rf").as_atom()),
            lineage_proof,
            [bytes32(ap.as_atom()) for ap in second_uncurried_cr_layer.args.at("rf").as_iter()],
            second_uncurried_cr_layer.args.at("rrf"),
            first_uncurried_cr_layer.args.at("rf").get_tree_hash(),
        )

    @classmethod
    def get_next_from_coin_spend(
        cls: type[_T_CRCAT],
        parent_spend: CoinSpend,
        conditions: Optional[Program] = None,  # For optimization purposes, the conditions may already have been run
    ) -> list[CRCAT]:
        """
        Given a coin spend, this will return the next CR-CATs that were created as an output of that spend.
        Inner puzzle output conditions may also be supplied as an optimization.

        This is the main method to use when syncing. It can also sync from a CAT spend that was not a CR-CAT so long
        as the spend output a remark condition that was (REMARK authorized_providers proofs_checker)
        """
        coin_name: bytes32 = parent_spend.coin.name()
        puzzle: Program = parent_spend.puzzle_reveal.to_program()
        solution: Program = parent_spend.solution.to_program()

        # Get info by uncurrying
        _, tail_hash_as_prog, potential_cr_layer = puzzle.uncurry()[1].as_iter()
        new_inner_puzzle_hash: Optional[bytes32] = None
        if potential_cr_layer.uncurry()[0].uncurry()[0] != CREDENTIAL_RESTRICTION:
            # If the previous spend is not a CR-CAT:
            # we look for a remark condition that tells us the authorized_providers and proofs_checker
            inner_solution: Program = solution.at("f")
            if conditions is None:
                conditions = potential_cr_layer.run(inner_solution)
            for condition in conditions.as_iter():
                if condition.at("f") == Program.to(1):
                    new_inner_puzzle_hash = bytes32(condition.at("rf").as_atom())
                    authorized_providers_as_prog: Program = condition.at("rrf")
                    proofs_checker: Program = condition.at("rrrf")
                    break
            else:
                raise ValueError(
                    "Previous spend was not a CR-CAT, nor did it properly remark the CR params"
                )  # pragma: no cover
            authorized_providers = [bytes32(p.as_atom()) for p in authorized_providers_as_prog.as_iter()]
            lineage_inner_puzhash: bytes32 = potential_cr_layer.get_tree_hash()
        else:
            # Otherwise the info we need will be in the puzzle reveal
            cr_first_curry, self_hash_and_innerpuz = potential_cr_layer.uncurry()
            _, authorized_providers_as_prog, proofs_checker = cr_first_curry.uncurry()[1].as_iter()
            _, inner_puzzle = self_hash_and_innerpuz.as_iter()
            inner_solution = solution.at("f").at("rrrrrrf")
            if conditions is None:
                conditions = inner_puzzle.run(inner_solution)
            inner_puzzle_hash: bytes32 = inner_puzzle.get_tree_hash()
            authorized_providers = [bytes32(p.as_atom()) for p in authorized_providers_as_prog.as_iter()]
            lineage_inner_puzhash = construct_cr_layer(
                authorized_providers,
                proofs_checker,
                inner_puzzle_hash,  # type: ignore
            ).get_tree_hash_precalc(inner_puzzle_hash)

        # Convert all of the old stuff into python
        new_lineage_proof: LineageProof = LineageProof(
            parent_spend.coin.parent_coin_info,
            lineage_inner_puzhash,
            uint64(parent_spend.coin.amount),
        )

        all_conditions: list[Program] = list(conditions.as_iter())
        if len(all_conditions) > 1000:
            raise RuntimeError("More than 1000 conditions not currently supported by CRCAT drivers")  # pragma: no cover

        # Almost complete except the coin's full puzzle hash which we want to use the class method to calculate
        partially_completed_crcats: list[CRCAT] = [
            CRCAT(
                Coin(coin_name, bytes(32), uint64(condition.at("rrf").as_int())),
                bytes32(tail_hash_as_prog.as_atom()),
                new_lineage_proof,
                authorized_providers,
                proofs_checker,
                bytes32(condition.at("rf").as_atom()) if new_inner_puzzle_hash is None else new_inner_puzzle_hash,
            )
            for condition in all_conditions
            if condition.at("f").as_int() == 51 and condition.at("rrf") != Program.to(-113)
        ]

        return [
            replace(
                crcat,
                coin=Coin(
                    crcat.coin.parent_coin_info,
                    crcat.construct_puzzle(crcat.inner_puzzle_hash).get_tree_hash_precalc(  # type: ignore
                        crcat.inner_puzzle_hash
                    ),
                    crcat.coin.amount,
                ),
            )
            for crcat in partially_completed_crcats
        ]

    def do_spend(
        self,
        # CAT solving info
        previous_coin_id: bytes32,
        next_coin_proof: LineageProof,
        previous_subtotal: int,
        extra_delta: int,
        # CR layer solving info
        proof_of_inclusions: Program,
        proof_checker_solution: Program,
        provider_id: bytes32,
        vc_launcher_id: bytes32,
        vc_inner_puzhash: Optional[bytes32],  # Optional for incomplete spends
        # Inner puzzle and solution
        inner_puzzle: Program,
        inner_solution: Program,
        # For optimization purposes the conditions may already have been run
        conditions: Optional[Iterable[Program]] = None,
    ) -> tuple[list[AssertCoinAnnouncement], CoinSpend, list[CRCAT]]:
        """
        Spend a CR-CAT.

        Must give the CAT accounting information, the valid VC proof, and the inner puzzle and solution.  The function
        will return the announcement IDs for the VC to optionally assert, the spend of this CAT, and the class
        representations of any CR-CAT outputs.

        Likely, spend_many is more useful.
        """
        # Gather the output information
        announcements: list[AssertCoinAnnouncement] = []
        new_inner_puzzle_hashes_and_amounts: list[tuple[bytes32, uint64]] = []
        if conditions is None:
            conditions = inner_puzzle.run(inner_solution).as_iter()  # pragma: no cover
        assert conditions is not None
        for condition in conditions:
            if condition.at("f").as_int() == 51 and condition.at("rrf").as_int() != -113:
                new_inner_puzzle_hash: bytes32 = bytes32(condition.at("rf").as_atom())
                new_amount: uint64 = uint64(condition.at("rrf").as_int())
                announcements.append(
                    AssertCoinAnnouncement(
                        asserted_id=self.coin.name(),
                        asserted_msg=b"\xcd" + std_hash(new_inner_puzzle_hash + int_to_bytes(new_amount)),
                    )
                )
                new_inner_puzzle_hashes_and_amounts.append((new_inner_puzzle_hash, new_amount))

        return (
            announcements,
            make_spend(
                self.coin,
                self.construct_puzzle(inner_puzzle),
                Program.to(  # solve_cat
                    [
                        solve_cr_layer(
                            proof_of_inclusions,
                            proof_checker_solution,
                            provider_id,
                            vc_launcher_id,
                            vc_inner_puzhash,  # type: ignore
                            self.coin.name(),
                            inner_solution,
                        ),
                        self.lineage_proof.to_program(),
                        previous_coin_id,
                        coin_as_list(self.coin),
                        next_coin_proof.to_program(),
                        previous_subtotal,
                        extra_delta,
                    ]
                ),
            ),
            [
                CRCAT(
                    Coin(
                        self.coin.name(),
                        self.construct_puzzle(new_inner_puzzle_hash).get_tree_hash_precalc(  # type: ignore
                            new_inner_puzzle_hash
                        ),
                        new_amount,
                    ),
                    self.tail_hash,
                    LineageProof(
                        self.coin.parent_coin_info,
                        self.construct_cr_layer(self.inner_puzzle_hash).get_tree_hash_precalc(  # type: ignore
                            self.inner_puzzle_hash
                        ),
                        uint64(self.coin.amount),
                    ),
                    self.authorized_providers,
                    self.proofs_checker,
                    new_inner_puzzle_hash,
                )
                for new_inner_puzzle_hash, new_amount in new_inner_puzzle_hashes_and_amounts
            ],
        )

    @classmethod
    def spend_many(
        cls: type[_T_CRCAT],
        inner_spends: list[tuple[_T_CRCAT, int, Program, Program]],  # CRCAT, extra_delta, inner puzzle, inner solution
        # CR layer solving info
        proof_of_inclusions: Program,
        proof_checker_solution: Program,
        provider_id: bytes32,
        vc_launcher_id: bytes32,
        vc_inner_puzhash: Optional[bytes32],  # Optional for incomplete spends
    ) -> tuple[list[AssertCoinAnnouncement], list[CoinSpend], list[CRCAT]]:
        """
        Spend a multiple CR-CATs.

        This class will handle all of the CAT accounting information, the only necessary information is the inner
        puzzle/solution, and the proof of a valid VC being spent along side all of the coins. There is currently no
        support for multiple VCs being used across the spend.  There is also currently no support for minting/melting.
        """

        def next_index(index: int) -> int:
            return 0 if index == len(inner_spends) - 1 else index + 1

        def prev_index(index: int) -> int:
            return index - 1

        sorted_inner_spends: list[tuple[_T_CRCAT, int, Program, Program]] = sorted(
            inner_spends,
            key=lambda spend: spend[0].coin.name(),
        )

        all_expected_announcements: list[AssertCoinAnnouncement] = []
        all_coin_spends: list[CoinSpend] = []
        all_new_crcats: list[CRCAT] = []

        subtotal: int = 0
        for i, inner_spend in enumerate(sorted_inner_spends):
            crcat, extra_delta, inner_puzzle, inner_solution = inner_spend
            conditions: list[Program] = list(inner_puzzle.run(inner_solution).as_iter())
            output_amount: int = (
                sum(
                    c.at("rrf").as_int()
                    for c in conditions
                    if c.at("f").as_int() == 51 and c.at("rrf").as_int() != -113
                )
                - extra_delta
            )
            next_crcat, _, _, _ = sorted_inner_spends[next_index(i)]
            prev_crcat, _, _, _ = sorted_inner_spends[prev_index(i)]
            expected_announcements, coin_spend, new_crcats = crcat.do_spend(
                prev_crcat.coin.name(),
                LineageProof(
                    next_crcat.coin.parent_coin_info,
                    next_crcat.construct_cr_layer(
                        next_crcat.inner_puzzle_hash,  # type: ignore
                    ).get_tree_hash_precalc(next_crcat.inner_puzzle_hash),
                    uint64(next_crcat.coin.amount),
                ),
                subtotal,
                extra_delta,
                proof_of_inclusions,
                proof_checker_solution,
                provider_id,
                vc_launcher_id,
                vc_inner_puzhash,
                inner_puzzle,
                inner_solution,
                conditions=conditions,
            )
            all_expected_announcements.extend(expected_announcements)
            all_coin_spends.append(coin_spend)
            all_new_crcats.extend(new_crcats)

            subtotal = subtotal + crcat.coin.amount - output_amount

        return all_expected_announcements, all_coin_spends, all_new_crcats

    def expected_announcement(self) -> bytes32:
        """
        The announcement a VC must make to this CAT in order to spend it
        """
        return std_hash(self.coin.name() + b"\xca")


@dataclass(frozen=True)
class CRCATSpend:
    crcat: CRCAT
    inner_puzzle: Program
    inner_solution: Program
    children: list[CRCAT]
    incomplete: bool
    inner_conditions: list[Program]
    proof_of_inclusions: Program

    @classmethod
    def from_coin_spend(cls, spend: CoinSpend) -> CRCATSpend:  # pragma: no cover
        inner_puzzle: Program = CRCAT.get_inner_puzzle(uncurry_puzzle(spend.puzzle_reveal))
        inner_solution: Program = CRCAT.get_inner_solution(spend.solution.to_program())
        inner_conditions: Program = inner_puzzle.run(inner_solution)
        return cls(
            CRCAT.get_current_from_coin_spend(spend),
            inner_puzzle,
            inner_solution,
            CRCAT.get_next_from_coin_spend(spend, conditions=inner_conditions),
            spend.solution.to_program().at("f").at("rrrrf") == Program.to(None),
            list(inner_conditions.as_iter()),
            spend.solution.to_program().at("f").at("f"),
        )


@streamable
@dataclass(frozen=True)
class ProofsChecker(Streamable):
    flags: list[str]

    def as_program(self) -> Program:
        def byte_sort_flags(f1: str, f2: str) -> int:
            return 1 if Program.to([10, (1, f1), (1, f2)]).run([]) == Program.to(None) else -1

        return PROOF_FLAGS_CHECKER.curry(
            [
                Program.to((flag, "1"))
                for flag in sorted(
                    self.flags,
                    key=functools.cmp_to_key(byte_sort_flags),
                )
            ]
        )

    @classmethod
    def from_program(cls, uncurried_puzzle: UncurriedPuzzle) -> ProofsChecker:
        if uncurried_puzzle.mod != PROOF_FLAGS_CHECKER:
            raise ValueError("Puzzle was not a proof checker")  # pragma: no cover

        return cls([flag.at("f").as_atom().decode("utf8") for flag in uncurried_puzzle.args.at("f").as_iter()])


class CRCATVersion(IntEnum):
    V1 = uint16(1)


@streamable
@dataclass(frozen=True)
class CRCATMetadata(Streamable):
    lineage_proof: LineageProof
    inner_puzzle_hash: bytes32



from __future__ import annotations

import asyncio
import logging
import random
from dataclasses import dataclass
from typing import Optional, cast

import pytest

from chik.full_node.tx_processing_queue import TransactionQueue, TransactionQueueFull
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.transaction_queue_entry import TransactionQueueEntry
from chik.util.task_referencer import create_referenced_task

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class FakeTransactionQueueEntry:
    index: int
    peer_id: Optional[bytes32]


def get_transaction_queue_entry(peer_id: Optional[bytes32], tx_index: int) -> TransactionQueueEntry:  # easy shortcut
    return cast(TransactionQueueEntry, FakeTransactionQueueEntry(index=tx_index, peer_id=peer_id))


@pytest.mark.anyio
async def test_local_txs(seeded_random: random.Random) -> None:
    transaction_queue = TransactionQueue(1000, log)
    # test 1 tx
    first_tx = get_transaction_queue_entry(None, 0)
    await transaction_queue.put(first_tx, None)

    result1 = await transaction_queue.pop()

    assert result1 == first_tx

    # test 2000 txs
    num_txs = 2000
    list_txs = [get_transaction_queue_entry(bytes32.random(seeded_random), i) for i in range(num_txs)]
    for tx in list_txs:
        await transaction_queue.put(tx, None)

    resulting_txs = []
    for _ in range(num_txs):
        resulting_txs.append(await transaction_queue.pop())

    for i in range(num_txs):
        assert list_txs[i] == resulting_txs[i]


@pytest.mark.anyio
async def test_one_peer_and_await(seeded_random: random.Random) -> None:
    transaction_queue = TransactionQueue(1000, log)
    num_txs = 100
    peer_id = bytes32.random(seeded_random)

    list_txs = [get_transaction_queue_entry(peer_id, i) for i in range(num_txs)]
    for tx in list_txs:
        await transaction_queue.put(tx, peer_id)

    # test transaction priority
    local_txs = [get_transaction_queue_entry(None, i) for i in range(int(num_txs / 5))]  # 20 txs
    for tx in local_txs:
        await transaction_queue.put(tx, None)

    resulting_txs = []
    for _ in range(num_txs + len(local_txs)):
        resulting_txs.append(await transaction_queue.pop())

    for i in range(num_txs + len(local_txs)):
        if i < len(local_txs):
            assert local_txs[i] == resulting_txs[i]  # first 20 should come from local
        else:
            assert list_txs[i - 20] == resulting_txs[i]

    # now we validate that the pop command is blocking
    task = create_referenced_task(transaction_queue.pop())
    with pytest.raises(asyncio.InvalidStateError):  # task is not done, so we expect an error when getting result
        task.result()
    # add a tx to test task completion
    await transaction_queue.put(get_transaction_queue_entry(None, 0), None)
    await asyncio.wait_for(task, 1)  # we should never time out here


@pytest.mark.anyio
async def test_lots_of_peers(seeded_random: random.Random) -> None:
    transaction_queue = TransactionQueue(1000, log)
    num_peers = 1000
    num_txs = 100
    total_txs = num_txs * num_peers
    peer_ids: list[bytes32] = [bytes32.random(seeded_random) for _ in range(num_peers)]

    # 100 txs per peer
    list_txs = [get_transaction_queue_entry(peer_id, i) for peer_id in peer_ids for i in range(num_txs)]
    for tx in list_txs:
        await transaction_queue.put(tx, tx.peer_id)  # type: ignore[attr-defined]

    resulting_txs = []
    for _ in range(total_txs):
        resulting_txs.append(await transaction_queue.pop())

    # There are 1000 peers, so each peer will have one transaction processed every 1000 iterations.
    for i in range(num_txs):
        assert list_txs[i] == resulting_txs[i * 1000]


@pytest.mark.anyio
async def test_full_queue(seeded_random: random.Random) -> None:
    transaction_queue = TransactionQueue(1000, log)
    num_peers = 100
    num_txs = 1000
    total_txs = num_txs * num_peers
    peer_ids: list[bytes32] = [bytes32.random(seeded_random) for _ in range(num_peers)]

    # 999 txs per peer then 1 to fail later
    list_txs = [get_transaction_queue_entry(peer_id, i) for peer_id in peer_ids for i in range(num_txs)]
    for tx in list_txs:
        await transaction_queue.put(tx, tx.peer_id)  # type: ignore[attr-defined]

    # test failure case.
    with pytest.raises(TransactionQueueFull):
        await transaction_queue.put(get_transaction_queue_entry(peer_ids[0], 1001), peer_ids[0])

    resulting_txs = []
    for _ in range(total_txs):
        resulting_txs.append(await transaction_queue.pop())


@pytest.mark.anyio
async def test_queue_cleanup_and_fairness(seeded_random: random.Random) -> None:
    transaction_queue = TransactionQueue(1000, log)
    peer_a = bytes32.random(seeded_random)
    peer_b = bytes32.random(seeded_random)
    peer_c = bytes32.random(seeded_random)

    # 2 for a, 1 for b, 2 for c
    peer_tx_a = [get_transaction_queue_entry(peer_a, i) for i in range(2)]
    peer_tx_b = [get_transaction_queue_entry(peer_b, 0)]
    peer_tx_c = [get_transaction_queue_entry(peer_c, i) for i in range(2)]

    list_txs = peer_tx_a + peer_tx_b + peer_tx_c
    for tx in list_txs:
        await transaction_queue.put(tx, tx.peer_id)  # type: ignore[attr-defined]

    resulting_ids = []
    for _ in range(3):  # we validate we get one transaction per peer
        resulting_ids.append((await transaction_queue.pop()).peer_id)  # type: ignore[attr-defined]
    assert [peer_a, peer_b, peer_c] == resulting_ids  # all peers have been properly included in the queue.
    second_resulting_ids = []
    for _ in range(2):  # we validate that we properly queue the last 2 transactions
        second_resulting_ids.append((await transaction_queue.pop()).peer_id)  # type: ignore[attr-defined]
    assert [peer_a, peer_c] == second_resulting_ids



from __future__ import annotations

import dataclasses
from collections.abc import Awaitable
from typing import Any, Callable, Optional

import pytest
from chik_rs import G2Element
from typing_extensions import Literal

from chik._tests.environments.wallet import WalletEnvironment, WalletStateTransition, WalletTestFramework
from chik._tests.util.time_out_assert import time_out_assert_not_none
from chik.rpc.wallet_request_types import VCAddProofs, VCGet, VCGetList, VCGetProofsForRoot, VCMint, VCRevoke, VCSpend
from chik.rpc.wallet_rpc_client import WalletRpcClient
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.types.blockchain_format.coin import coin_as_list
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.peer_info import PeerInfo
from chik.util.bech32m import encode_puzzle_hash
from chik.util.ints import uint64
from chik.wallet.cat_wallet.cat_utils import CAT_MOD, construct_cat_puzzle
from chik.wallet.cat_wallet.cat_wallet import CATWallet
from chik.wallet.did_wallet.did_wallet import DIDWallet
from chik.wallet.util.query_filter import TransactionTypeFilter
from chik.wallet.util.transaction_type import TransactionType
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.vc_wallet.cr_cat_drivers import ProofsChecker, construct_cr_layer
from chik.wallet.vc_wallet.cr_cat_wallet import CRCATWallet
from chik.wallet.vc_wallet.vc_store import VCProofs, VCRecord
from chik.wallet.wallet import Wallet
from chik.wallet.wallet_node import WalletNode
from chik.wallet.wallet_spend_bundle import WalletSpendBundle


async def mint_cr_cat(
    num_blocks: int,
    wallet_0: Wallet,
    wallet_node_0: WalletNode,
    client_0: WalletRpcClient,
    full_node_api: FullNodeSimulator,
    authorized_providers: list[bytes32] = [],
    tail: Program = Program.to(None),
    proofs_checker: ProofsChecker = ProofsChecker(["foo", "bar"]),
) -> None:
    our_puzzle: Program = await wallet_0.get_new_puzzle()
    cat_puzzle: Program = construct_cat_puzzle(
        CAT_MOD,
        tail.get_tree_hash(),
        Program.to(1),
    )
    CAT_AMOUNT_0 = uint64(100)

    await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_0, timeout=20)
    tx = (
        await client_0.create_signed_transactions(
            [
                {
                    "puzzle_hash": cat_puzzle.get_tree_hash(),
                    "amount": CAT_AMOUNT_0,
                }
            ],
            DEFAULT_TX_CONFIG,
            wallet_id=1,
        )
    ).signed_tx
    spend_bundle = tx.spend_bundle
    assert spend_bundle is not None

    # Do the eve spend back to our wallet and add the CR layer
    cat_coin = next(c for c in spend_bundle.additions() if c.amount == CAT_AMOUNT_0)
    eve_spend = WalletSpendBundle(
        [
            make_spend(
                cat_coin,
                cat_puzzle,
                Program.to(
                    [
                        Program.to(
                            [
                                [
                                    51,
                                    construct_cr_layer(
                                        authorized_providers,
                                        proofs_checker.as_program(),
                                        our_puzzle,
                                    ).get_tree_hash(),
                                    CAT_AMOUNT_0,
                                    [our_puzzle.get_tree_hash()],
                                ],
                                [51, None, -113, tail, None],
                                [1, our_puzzle.get_tree_hash(), authorized_providers, proofs_checker.as_program()],
                            ]
                        ),
                        None,
                        cat_coin.name(),
                        coin_as_list(cat_coin),
                        [cat_coin.parent_coin_info, Program.to(1).get_tree_hash(), cat_coin.amount],
                        0,
                        0,
                    ]
                ),
            )
        ],
        G2Element(),
    )
    spend_bundle = WalletSpendBundle.aggregate([spend_bundle, eve_spend])
    await wallet_node_0.wallet_state_manager.add_pending_transactions(
        [dataclasses.replace(tx, spend_bundle=spend_bundle, name=spend_bundle.name())]
    )
    await time_out_assert_not_none(5, full_node_api.full_node.mempool_manager.get_spendbundle, spend_bundle.name())


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 2,
            "config_overrides": {"automatically_add_unknown_cats": True},
            "blocks_needed": [2, 1],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_vc_lifecycle(wallet_environments: WalletTestFramework) -> None:
    # Setup
    full_node_api: FullNodeSimulator = wallet_environments.full_node
    env_0 = wallet_environments.environments[0]
    env_1 = wallet_environments.environments[1]
    wallet_node_0 = wallet_environments.environments[0].node
    wallet_node_1 = wallet_environments.environments[1].node
    wallet_0 = wallet_environments.environments[0].xck_wallet
    wallet_1 = wallet_environments.environments[1].xck_wallet
    client_0 = wallet_environments.environments[0].rpc_client
    client_1 = wallet_environments.environments[1].rpc_client

    # Define wallet aliases
    env_0.wallet_aliases = {
        "xck": 1,
        "did": 2,
        "vc": 3,
        "crcat": 4,
    }
    env_1.wallet_aliases = {
        "xck": 1,
        "crcat": 2,
        "vc": 3,
    }

    # Generate DID as an "authorized provider"
    async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_id: bytes32 = bytes32.from_hexstr(
            (
                await DIDWallet.create_new_did_wallet(
                    wallet_node_0.wallet_state_manager, wallet_0, uint64(1), action_scope
                )
            ).get_my_DID()
        )

    # Mint a VC
    vc_record = (
        await client_0.vc_mint(
            VCMint(
                did_id=encode_puzzle_hash(did_id, "did"),
                target_address=encode_puzzle_hash(await wallet_0.get_new_puzzlehash(), "txck"),
                fee=uint64(1_750_000_000_000),
                push=True,
            ),
            wallet_environments.tx_config,
        )
    ).vc_record

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        # 1_750_000_000_000 for VC mint fee, 1 for VC singleton, 1 for DID mint
                        "unconfirmed_wallet_balance": -1_750_000_000_002,
                        # I'm not sure incrementing pending_coin_removal_count here by 3 is the spirit of this number
                        # One existing coin has been removed and two ephemeral coins have been removed
                        # Does pending_coin_removal_count attempt to show the number of current pending removals
                        # Or does it intend to just mean all pending removals that we should eventually get states for?
                        "pending_coin_removal_count": 3,
                        "<=#spendable_balance": -1_750_000_000_002,
                        "<=#max_send_amount": -1_750_000_000_002,
                        "set_remainder": True,
                    },
                    "did": {"init": True, "set_remainder": True},
                    "vc": {
                        "init": True,
                        "confirmed_wallet_balance": 0,
                        "unconfirmed_wallet_balance": 0,
                        "spendable_balance": 0,
                        "pending_change": 0,
                        "max_send_amount": 0,
                        "unspent_coin_count": 0,
                        "pending_coin_removal_count": 0,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        # 1_750_000_000_000 for VC mint fee, 1 for VC singleton, 1 for DID mint
                        "confirmed_wallet_balance": -1_750_000_000_002,
                        "pending_coin_removal_count": -3,  # 3 for VC mint, 1 for DID mint
                        "set_remainder": True,
                    },
                    "did": {
                        "set_remainder": True,
                    },
                    "vc": {
                        "unspent_coin_count": 1,
                    },
                },
            ),
            WalletStateTransition(),
        ]
    )
    new_vc_record: Optional[VCRecord] = (await client_0.vc_get(VCGet(vc_record.vc.launcher_id))).vc_record
    assert new_vc_record is not None

    # Spend VC
    proofs: VCProofs = VCProofs({"foo": "1", "bar": "1", "baz": "1", "qux": "1", "grault": "1"})
    proof_root: bytes32 = proofs.root()
    await client_0.vc_spend(
        VCSpend(
            vc_id=vc_record.vc.launcher_id,
            new_proof_hash=proof_root,
            fee=uint64(100),
            push=True,
        ),
        wallet_environments.tx_config,
    )
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": -100,
                        "pending_coin_removal_count": 1,
                        "<=#spendable_balance": -100,
                        "<=#max_send_amount": -100,
                        "set_remainder": True,
                    },
                    "did": {
                        "spendable_balance": -1,
                        "pending_change": 1,
                        "pending_coin_removal_count": 1,
                        "max_send_amount": -1,
                    },
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": -100,
                        "pending_coin_removal_count": -1,
                        "set_remainder": True,
                    },
                    "did": {
                        "spendable_balance": 1,
                        "pending_change": -1,
                        "pending_coin_removal_count": -1,
                        "max_send_amount": 1,
                    },
                    "vc": {
                        "pending_coin_removal_count": -1,
                    },
                },
            ),
            WalletStateTransition(),
        ]
    )
    vc_record_updated: Optional[VCRecord] = (await client_0.vc_get(VCGet(vc_record.vc.launcher_id))).vc_record
    assert vc_record_updated is not None
    assert vc_record_updated.vc.proof_hash == proof_root

    # Do a mundane spend
    await client_0.vc_spend(VCSpend(vc_id=vc_record.vc.launcher_id, push=True), wallet_environments.tx_config)
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": -1,
                    },
                },
            ),
            WalletStateTransition(),
        ]
    )

    # Add proofs to DB
    await client_0.vc_add_proofs(VCAddProofs.from_vc_proofs(proofs))
    # Doing it again just to make sure it doesn't care
    await client_0.vc_add_proofs(VCAddProofs.from_vc_proofs(proofs))
    assert (
        await client_0.vc_get_proofs_for_root(VCGetProofsForRoot(proof_root))
    ).to_vc_proofs().key_value_pairs == proofs.key_value_pairs
    get_list_reponse = await client_0.vc_get_list(VCGetList())
    assert len(get_list_reponse.vc_records) == 1
    assert get_list_reponse.proof_dict[proof_root] == proofs.key_value_pairs

    # Mint CR-CAT
    await mint_cr_cat(1, wallet_0, wallet_node_0, client_0, full_node_api, [did_id])
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": -100,
                        "<=#spendable_balance": -100,
                        "<=#max_send_amount": -100,
                        "set_remainder": True,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": -100,
                        "set_remainder": True,
                    },
                    "crcat": {
                        "init": True,
                        "confirmed_wallet_balance": 100,
                        "unconfirmed_wallet_balance": 100,
                        "spendable_balance": 100,
                        "pending_change": 0,
                        "max_send_amount": 100,
                        "unspent_coin_count": 1,
                        "pending_coin_removal_count": 0,
                    },
                },
            ),
            WalletStateTransition(),
        ]
    )

    cr_cat_wallet_0 = wallet_node_0.wallet_state_manager.wallets[env_0.dealias_wallet_id("crcat")]
    assert isinstance(cr_cat_wallet_0, CRCATWallet)
    assert await CRCATWallet.create(  # just testing the create method doesn't throw
        wallet_node_0.wallet_state_manager,
        wallet_node_0.wallet_state_manager.main_wallet,
        (await wallet_node_0.wallet_state_manager.get_all_wallet_info_entries(wallet_type=WalletType.CRCAT))[0],
    )
    assert {
        "data": bytes(cr_cat_wallet_0.info).hex(),
        "id": env_0.dealias_wallet_id("crcat"),
        "name": cr_cat_wallet_0.get_name(),
        "type": cr_cat_wallet_0.type(),
        "authorized_providers": [p.hex() for p in cr_cat_wallet_0.info.authorized_providers],
        "flags_needed": cr_cat_wallet_0.info.proofs_checker.flags,
    } == (await client_0.get_wallets(wallet_type=cr_cat_wallet_0.type()))[0]
    assert await wallet_node_0.wallet_state_manager.get_wallet_for_asset_id(cr_cat_wallet_0.get_asset_id()) is not None
    wallet_1_ph = await wallet_1.get_new_puzzlehash()
    wallet_1_addr = encode_puzzle_hash(wallet_1_ph, "txck")
    txs = (
        await client_0.cat_spend(
            cr_cat_wallet_0.id(),
            wallet_environments.tx_config,
            uint64(90),
            wallet_1_addr,
            uint64(2000000000),
            memos=["hey"],
        )
    ).transactions
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": -2000000000,
                        "pending_coin_removal_count": 1,
                        "<=#spendable_balance": -2000000000,
                        "<=#max_send_amount": -2000000000,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                    "crcat": {
                        "unconfirmed_wallet_balance": -90,
                        "spendable_balance": -100,
                        "max_send_amount": -100,
                        "pending_change": 10,
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": -2000000000,
                        "pending_coin_removal_count": -1,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": -1,
                    },
                    "crcat": {
                        "confirmed_wallet_balance": -90,
                        "spendable_balance": 10,
                        "max_send_amount": 10,
                        "pending_change": -10,
                        "pending_coin_removal_count": -1,
                    },
                },
            ),
            WalletStateTransition(
                post_block_balance_updates={
                    "crcat": {
                        "init": True,
                        "confirmed_wallet_balance": 0,
                        "unconfirmed_wallet_balance": 0,
                        "spendable_balance": 0,
                        "pending_change": 0,
                        "max_send_amount": 0,
                        "unspent_coin_count": 0,
                        "pending_coin_removal_count": 0,
                    }
                },
                post_block_additional_balance_info={
                    "crcat": {
                        "pending_approval_balance": 90,
                    },
                },
            ),
        ]
    )
    assert await wallet_node_1.wallet_state_manager.wallets[env_1.dealias_wallet_id("crcat")].match_hinted_coin(
        next(c for tx in txs for c in tx.additions if c.amount == 90), wallet_1_ph
    )
    pending_tx = await client_1.get_transactions(
        env_1.dealias_wallet_id("crcat"),
        0,
        1,
        reverse=True,
        type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_CRCAT_PENDING]),
    )
    assert len(pending_tx) == 1

    # Send the VC to wallet_1 to use for the CR-CATs
    await client_0.vc_spend(
        VCSpend(vc_id=vc_record.vc.launcher_id, new_puzhash=await wallet_1.get_new_puzzlehash(), push=True),
        wallet_environments.tx_config,
    )
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": 1,
                    }
                },
                post_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": -1,
                        "unspent_coin_count": -1,
                    }
                },
            ),
            WalletStateTransition(
                post_block_balance_updates={
                    "vc": {"init": True, "set_remainder": True},
                }
            ),
        ]
    )
    await client_1.vc_add_proofs(VCAddProofs.from_vc_proofs(proofs))

    # Claim the pending approval to our wallet
    await client_1.crcat_approve_pending(
        env_1.dealias_wallet_id("crcat"),
        uint64(90),
        wallet_environments.tx_config,
        fee=uint64(90),
    )
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(),
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": -90,
                        "pending_coin_removal_count": 1,
                        "<=#spendable_balance": -90,
                        "<=#max_send_amount": -90,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                    "crcat": {
                        "unconfirmed_wallet_balance": 90,
                        "pending_change": 90,
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": -90,
                        "pending_coin_removal_count": -1,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": -1,
                    },
                    "crcat": {
                        "confirmed_wallet_balance": 90,
                        "spendable_balance": 90,
                        "max_send_amount": 90,
                        "pending_change": -90,
                        "unspent_coin_count": 1,
                        "pending_coin_removal_count": -1,
                    },
                },
                post_block_additional_balance_info={
                    "crcat": {
                        "pending_approval_balance": 0,
                    },
                },
            ),
        ]
    )

    # (Negative test) Try to spend a CR-CAT that we don't have a valid VC for
    with pytest.raises(ValueError):
        await client_0.cat_spend(
            cr_cat_wallet_0.id(),
            wallet_environments.tx_config,
            uint64(10),
            wallet_1_addr,
        )

    # Test melting a CRCAT
    # This is intended to trigger an edge case where the output and change are the same forcing a new puzhash
    with wallet_environments.new_puzzle_hashes_allowed():
        tx = (
            await client_1.cat_spend(
                env_1.dealias_wallet_id("crcat"),
                wallet_environments.tx_config,
                uint64(20),
                wallet_1_addr,
                uint64(0),
                cat_discrepancy=(-50, Program.to(None), Program.to(None)),
            )
        ).transaction
    [tx] = await wallet_node_1.wallet_state_manager.add_pending_transactions([tx])
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(),
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": 20,
                        "pending_coin_removal_count": 1,
                        "<=#spendable_balance": 20,
                        "<=#max_send_amount": 20,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                    "crcat": {
                        "unconfirmed_wallet_balance": -50,
                        "spendable_balance": -90,
                        "max_send_amount": -90,
                        "pending_change": 40,
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": 20,
                        "pending_coin_removal_count": -1,
                        "set_remainder": True,
                    },
                    "vc": {
                        "pending_coin_removal_count": -1,
                    },
                    "crcat": {
                        "confirmed_wallet_balance": -50,  # should go straight to confirmed because we sent to ourselves
                        "spendable_balance": 40,
                        "max_send_amount": 40,
                        "pending_change": -40,
                        "pending_coin_removal_count": -1,
                        "unspent_coin_count": 1,
                    },
                },
            ),
        ]
    )
    vc_record_updated = (await client_1.vc_get(VCGet(vc_record_updated.vc.launcher_id))).vc_record
    assert vc_record_updated is not None

    # Revoke VC
    await client_0.vc_revoke(
        VCRevoke(vc_parent_id=vc_record_updated.vc.coin.parent_coin_info, fee=uint64(1), push=True),
        wallet_environments.tx_config,
    )
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "unconfirmed_wallet_balance": -1,
                        "pending_coin_removal_count": 1,
                        "<=#spendable_balance": -1,
                        "<=#max_send_amount": -1,
                        "set_remainder": True,
                    },
                    "did": {
                        "spendable_balance": -1,
                        "pending_change": 1,
                        "pending_coin_removal_count": 1,
                        "max_send_amount": -1,
                    },
                },
                post_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": -1,
                        "pending_coin_removal_count": -1,
                        "set_remainder": True,
                    },
                    "did": {
                        "spendable_balance": 1,
                        "pending_change": -1,
                        "pending_coin_removal_count": -1,
                        "max_send_amount": 1,
                    },
                },
            ),
            WalletStateTransition(
                post_block_balance_updates={
                    "vc": {
                        "unspent_coin_count": -1,
                    },
                },
            ),
        ]
    )
    assert (
        len(await (await wallet_node_0.wallet_state_manager.get_or_create_vc_wallet()).store.get_unconfirmed_vcs()) == 0
    )


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_self_revoke(wallet_environments: WalletTestFramework) -> None:
    # Setup
    env_0: WalletEnvironment = wallet_environments.environments[0]
    wallet_node_0 = env_0.node
    wallet_0 = env_0.xck_wallet
    client_0 = env_0.rpc_client

    # Aliases
    env_0.wallet_aliases = {
        "xck": 1,
        "did": 2,
        "vc": 3,
    }

    # Generate DID as an "authorized provider"
    async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node_0.wallet_state_manager, wallet_0, uint64(1), action_scope
        )
    did_id: bytes32 = bytes32.from_hexstr(did_wallet.get_my_DID())

    vc_record = (
        await client_0.vc_mint(
            VCMint(
                did_id=encode_puzzle_hash(did_id, "did"),
                target_address=encode_puzzle_hash(await wallet_0.get_new_puzzlehash(), "txck"),
                fee=uint64(200),
                push=True,
            ),
            wallet_environments.tx_config,
        )
    ).vc_record
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                # Balance checking for this spend covered in test_vc_lifecycle
                pre_block_balance_updates={
                    "xck": {"set_remainder": True},
                    "did": {"init": True, "set_remainder": True},
                    "vc": {"init": True, "set_remainder": True},
                },
                post_block_balance_updates={
                    "xck": {"set_remainder": True},
                    "did": {"set_remainder": True},
                    "vc": {"set_remainder": True},
                },
            )
        ]
    )
    new_vc_record: Optional[VCRecord] = (await client_0.vc_get(VCGet(vc_record.vc.launcher_id))).vc_record
    assert new_vc_record is not None

    # Test a negative case real quick (mostly unrelated)
    with pytest.raises(ValueError, match="at the same time"):
        async with wallet_node_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=False
        ) as action_scope:
            await (await wallet_node_0.wallet_state_manager.get_or_create_vc_wallet()).generate_signed_transaction(
                new_vc_record.vc.launcher_id,
                action_scope,
                new_proof_hash=bytes32.zeros,
                self_revoke=True,
            )

    # Send the DID to oblivion
    async with did_wallet.wallet_state_manager.new_action_scope(
        wallet_environments.tx_config, push=True
    ) as action_scope:
        await did_wallet.transfer_did(bytes32.zeros, uint64(0), False, action_scope)

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "did": {"set_remainder": True},
                },
                post_block_balance_updates={},
            )
        ]
    )

    # Make sure revoking still works
    await client_0.vc_revoke(
        VCRevoke(vc_parent_id=new_vc_record.vc.coin.parent_coin_info, push=True), wallet_environments.tx_config
    )
    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                # Balance checking for this spend covered in test_vc_lifecycle
                pre_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    "vc": {
                        "pending_coin_removal_count": -1,
                        "unspent_coin_count": -1,
                    },
                },
            )
        ]
    )
    vc_record_revoked: Optional[VCRecord] = (await client_0.vc_get(VCGet(vc_record.vc.launcher_id))).vc_record
    assert vc_record_revoked is None
    assert (
        len(await (await wallet_node_0.wallet_state_manager.get_or_create_vc_wallet()).store.get_unconfirmed_vcs()) == 0
    )


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_cat_wallet_conversion(
    self_hostname: str,
    one_wallet_and_one_simulator_services: Any,
    trusted: Any,
) -> None:
    num_blocks = 1
    full_nodes, wallets, bt = one_wallet_and_one_simulator_services
    full_node_api: FullNodeSimulator = full_nodes[0]._api
    full_node_server = full_node_api.full_node.server
    wallet_service_0 = wallets[0]
    wallet_node_0 = wallet_service_0._node
    wallet_0 = wallet_node_0.wallet_state_manager.main_wallet

    client_0 = await WalletRpcClient.create(
        bt.config["self_hostname"],
        wallet_service_0.rpc_server.listen_port,
        wallet_service_0.root_path,
        wallet_service_0.config,
    )

    if trusted:
        wallet_node_0.config["trusted_peers"] = {
            full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
        }
    else:
        wallet_node_0.config["trusted_peers"] = {}

    await wallet_node_0.server.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
    await full_node_api.farm_blocks_to_wallet(count=num_blocks, wallet=wallet_0)
    await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_0, timeout=20)

    # Key point of test: create a normal CAT wallet first, and see if it gets converted to CR-CAT wallet
    await CATWallet.get_or_create_wallet_for_cat(
        wallet_node_0.wallet_state_manager, wallet_0, Program.to(None).get_tree_hash().hex()
    )

    did_id = bytes32.zeros
    await mint_cr_cat(num_blocks, wallet_0, wallet_node_0, client_0, full_node_api, [did_id])
    await full_node_api.farm_blocks_to_wallet(count=num_blocks, wallet=wallet_0)
    await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_0, timeout=20)

    async def check_length(length: int, func: Callable[..., Awaitable[Any]], *args: Any) -> Optional[Literal[True]]:
        if len(await func(*args)) == length:
            return True
        return None  # pragma: no cover

    await time_out_assert_not_none(
        15, check_length, 1, wallet_node_0.wallet_state_manager.get_all_wallet_info_entries, WalletType.CRCAT
    )
    await time_out_assert_not_none(
        15, check_length, 0, wallet_node_0.wallet_state_manager.get_all_wallet_info_entries, WalletType.CAT
    )
